# PSyCO
# Copyright (C) 2014 Nuno Lopes
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from z3 import *
import conf

def is_int(n):
  return isinstance(n, (long, int))


class BaseClass:
  def __repr__(self):
    raise Exception('tostr not implemented')

  # we need structural comparison and not pointer comparison since our objects
  # are not uniqued
  def __eq__(self, other):
    return self.__class__ == other.__class__ and self.__dict__ == other.__dict__

  def __hash__(self):
    h = hash(self.__class__)
    for k,v in self.__dict__.iteritems():
      h ^= hash(v)
    return h


################################
class Expr(BaseClass):
  pass

class Num(Expr):
  def __init__(self, n):
    self.n = n
    assert(is_int(self.n))

  def __repr__(self):
    return str(self.n)

  def add_var_offset(self, off):
    return

  def collect_tmpls(self):
    return set()

  def collect_vars(self):
    return set()

  def to_z3_uf(self, vmap, rset, wset):
    return IntVal(self.n)

  def to_z3_nouf(self, vmap, tmap, rset, wset):
    return IntVal(self.n)

def is_num(n):
  return isinstance(n, Num)


class ExpT(Expr):
  def __init__(self, n):
    self.n = n
    assert(is_int(self.n))

  def __repr__(self):
    return 'E' + str(self.n)

  def collect_tmpls(self):
    return set([self])

  def collect_vars(self):
    return set()

  def to_z3_uf(self, vmap, rset, wset):
    assert wset[self] == []
    rvars = [z3.If(Bool('R_%s_%s' % (self, v)), vmap[v], 0) for v in rset[self]]
    fn_sort = [IntSort() for i in range(len(rvars)+1)]
    return Function(str(self), fn_sort)(rvars)

  def to_z3_nouf(self, vmap, tmap, rset, wset):
    assert wset[self] == []
    rvars = [vmap[v] for v in rset[self]]
    var = FreshInt('%s_' % self)
    if not tmap.has_key((self, None)):
        tmap[(self, None)] = []
    tmap[(self, None)] += [(var, rvars)]
    return var


class BinOp(Expr):
  Add, Sub, Mul, Div, Last = range(5)

  def __init__(self, e1, e2, op):
    self.e1 = e1
    self.e2 = e2
    self.op = op
    assert(isinstance(self.e1, Expr))
    assert(isinstance(self.e2, Expr))
    assert(self.op >= 0 and self.op < self.Last)

  def op_str(self):
    return {
      self.Add: '+',
      self.Sub: '-',
      self.Mul: '*',
      self.Div: '/',
      }[self.op]

  def __repr__(self):
    return '(' + str(self.e1) + ' ' + self.op_str() + ' ' + str(self.e2) + ')'

  def collect_tmpls(self):
    return self.e1.collect_tmpls() | self.e2.collect_tmpls()

  def collect_vars(self):
    return self.e1.collect_vars() | self.e2.collect_vars()

  def to_z3_uf(self, vmap, rset, wset):
    e1 = self.e1.to_z3_uf(vmap, rset, wset)
    e2 = self.e2.to_z3_uf(vmap, rset, wset)
    return {
      self.Add: e1 + e2,
      self.Sub: e1 - e2,
      self.Mul: e1 * e2,
      self.Div: e1 / e2,
      }[self.op]

  def to_z3_nouf(self, vmap, tmap, rset, wset):
    e1 = self.e1.to_z3_nouf(vmap, tmap, rset, wset)
    e2 = self.e2.to_z3_nouf(vmap, tmap, rset, wset)
    return {
      self.Add: e1 + e2,
      self.Sub: e1 - e2,
      self.Mul: e1 * e2,
      self.Div: e1 / e2,
      }[self.op]


class Min(Expr):
  def __init__(self, e1, e2):
    self.e1 = e1
    self.e2 = e2
    assert(isinstance(self.e1, Expr))
    assert(isinstance(self.e2, Expr))

  def __repr__(self):
    return 'min(' + str(self.e1) + ', ' + str(self.e2) + ')'

  def collect_tmpls(self):
    return self.e1.collect_tmpls() | self.e2.collect_tmpls()

  def collect_vars(self):
    return self.e1.collect_vars() | self.e2.collect_vars()

  def to_z3_uf(self, vmap, rset, wset):
    e1 = self.e1.to_z3_uf(vmap, rset, wset)
    e2 = self.e2.to_z3_uf(vmap, rset, wset)
    return z3.If(e1 < e2, e1, e2)

  def to_z3_nouf(self, vmap, tmap, rset, wset):
    e1 = self.e1.to_z3_nouf(vmap, tmap, rset, wset)
    e2 = self.e2.to_z3_nouf(vmap, tmap, rset, wset)
    return z3.If(e1 < e2, e1, e2)


class Max(Expr):
  def __init__(self, e1, e2):
    self.e1 = e1
    self.e2 = e2
    assert(isinstance(self.e1, Expr))
    assert(isinstance(self.e2, Expr))

  def __repr__(self):
    return 'max(' + str(self.e1) + ', ' + str(self.e2) + ')'

  def collect_tmpls(self):
    return self.e1.collect_tmpls() | self.e2.collect_tmpls()

  def collect_vars(self):
    return self.e1.collect_vars() | self.e2.collect_vars()

  def to_z3_uf(self, vmap, rset, wset):
    e1 = self.e1.to_z3_uf(vmap, rset, wset)
    e2 = self.e2.to_z3_uf(vmap, rset, wset)
    return z3.If(e1 > e2, e1, e2)

  def to_z3_nouf(self, vmap, tmap, rset, wset):
    e1 = self.e1.to_z3_nouf(vmap, tmap, rset, wset)
    e2 = self.e2.to_z3_nouf(vmap, tmap, rset, wset)
    return z3.If(e1 > e2, e1, e2)


################################
class Lval(Expr):
  pass


class VarL(Lval):
  def __init__(self, n):
    self.n = n
    assert(is_int(self.n))

  def __repr__(self):
    return "V" + str(self.n)

  def add_var_offset(self, off):
    self.n += off

  def collect_tmpls(self):
    return set()

  def collect_vars(self):
    return set([self])

  def to_z3_uf(self, vmap, rset, wset):
    return vmap[self]

  def to_z3_nouf(self, vmap, tmap, rset, wset):
    return vmap[self]


################################
class BoolExpr(BaseClass):
  pass


class TrueExpr(BoolExpr):
  def __repr__(self):
    return "true"

  def neg(self):
    return FalseExpr()

  def add_var_offset(self, off):
    return

  def collect_tmpls(self):
    return set()

  def collect_vars(self):
    return set()

  def to_z3_uf(self, vmap, rset, wset):
    return BoolVal(True)

  def to_z3_nouf(self, vmap, tmap, rset, wset):
    return BoolVal(True)


class FalseExpr(BoolExpr):
  def __repr__(self):
    return "false"

  def neg(self):
    return TrueExpr()

  def add_var_offset(self, off):
    return

  def collect_tmpls(self):
    return set()

  def collect_vars(self):
    return set()

  def to_z3_uf(self, vmap, rset, wset):
    return BoolVal(False)

  def to_z3_nouf(self, vmap, tmap, rset, wset):
    return BoolVal(False)


class BoolExpT(BoolExpr):
  def __init__(self, n):
    self.n = n
    assert(is_int(self.n))

  def __repr__(self):
    return 'B' + str(self.n)

  def neg(self):
    return NotExpr(self)

  def collect_tmpls(self):
    return set([self])

  def collect_vars(self):
    return set()

  def to_z3_uf(self, vmap, rset, wset):
    assert wset[self] == []
    rvars = [z3.If(Bool('R_%s_%s' % (self, v)), vmap[v], 0) for v in rset[self]]
    fn_sort = [IntSort() for i in range(len(rvars))] + [BoolSort()]
    return Function(str(self), fn_sort)(rvars)

  def to_z3_nouf(self, vmap, tmap, rset, wset):
    assert wset[self] == []
    rvars = [vmap[v] for v in rset[self]]
    var = FreshBool('%s_' % self)
    if not tmap.has_key((self, None)):
        tmap[(self, None)] = []
    tmap[(self, None)] += [(var, rvars)]
    return var

def is_boolexpt(p):
  return isinstance(p, BoolExpT)


class NotExpr(BoolExpr):
  def __init__(self, e):
    self.e = e
    assert(isinstance(self.e, BoolExpr))

  def __repr__(self):
    return '!(' + str(self.e) + ')'

  def neg(self):
    return self.e

  def collect_tmpls(self):
    return self.e.collect_tmpls()

  def collect_vars(self):
    return self.e.collect_vars()

  def to_z3_uf(self, vmap, rset, wset):
    return Not(self.e.to_z3_uf(vmap, rset, wset))

  def to_z3_nouf(self, vmap, tmap, rset, wset):
    return Not(self.e.to_z3_nouf(vmap, tmap, rset, wset))


class AndExpr(BoolExpr):
  def __init__(self, e1, e2):
    self.e1 = e1
    self.e2 = e2
    assert(isinstance(self.e1, BoolExpr))
    assert(isinstance(self.e2, BoolExpr))

  def __repr__(self):
    return str(self.e1) + ' && ' + str(self.e2)

  def neg(self):
    return OrExpr(self.e1.neg(), self.e2.neg())

  def collect_tmpls(self):
    return self.e1.collect_tmpls() | self.e2.collect_tmpls()

  def collect_vars(self):
    return self.e1.collect_vars() | self.e2.collect_vars()

  def to_z3_uf(self, vmap, rset, wset):
    return And(self.e1.to_z3_uf(vmap, rset, wset),
               self.e2.to_z3_uf(vmap, rset, wset))

  def to_z3_uf(self, vmap, tmap, rset, wset):
    return And(self.e1.to_z3_nouf(vmap, tmap, rset, wset),
               self.e2.to_z3_nouf(vmap, tmap, rset, wset))


class OrExpr(BoolExpr):
  def __init__(self, e1, e2):
    self.e1 = e1
    self.e2 = e2
    assert(isinstance(self.e1, BoolExpr))
    assert(isinstance(self.e2, BoolExpr))

  def __repr__(self):
    return str(self.e1) + ' || ' + str(self.e2)

  def neg(self):
    return OrExpr(self.e1.neg(), self.e2.neg())

  def collect_tmpls(self):
    return self.e1.collect_tmpls() | self.e2.collect_tmpls()

  def collect_vars(self):
    return self.e1.collect_vars() | self.e2.collect_vars()

  def to_z3_uf(self, vmap, rset, wset):
    return Or(self.e1.to_z3_uf(vmap, rset, wset),
              self.e2.to_z3_uf(vmap, rset, wset))

  def to_z3_nouf(self, vmap, tmap, rset, wset):
    return Or(self.e1.to_z3_nouf(vmap, tmap, rset, wset),
              self.e2.to_z3_nouf(vmap, tmap, rset, wset))


class CmpOp(BoolExpr):
  LT, LE, GT, GE, EQ, NE, Last = range(7)

  def __init__(self, e1, e2, op):
    self.e1 = e1
    self.e2 = e2
    self.op = op
    assert(isinstance(self.e1, Expr))
    assert(isinstance(self.e2, Expr))
    assert(self.op >= 0 and self.op < self.Last)

  def op_str(self):
    return {
      self.LT: '<',
      self.LE: '<=',
      self.GT: '>',
      self.GE: '>=',
      self.EQ: '=',
      self.NE: '!=',
      }[self.op]

  def __repr__(self):
    return str(self.e1) + ' ' + self.op_str() + ' ' + str(self.e2)

  def neg(self):
    return {
      self.LT: CmpOp(self.e1, self.e2, self.GE),
      self.LE: CmpOp(self.e1, self.e2, self.GT),
      self.GT: CmpOp(self.e1, self.e2, self.LE),
      self.GE: CmpOp(self.e1, self.e2, self.LT),
      self.EQ: CmpOp(self.e1, self.e2, self.NE),
      self.NE: CmpOp(self.e1, self.e2, self.EQ),
      }[self.op]

  def collect_tmpls(self):
    return self.e1.collect_tmpls() | self.e2.collect_tmpls()

  def collect_vars(self):
    return self.e1.collect_vars() | self.e2.collect_vars()

  def to_z3_uf(self, vmap, rset, wset):
    e1 = self.e1.to_z3_uf(vmap, rset, wset)
    e2 = self.e2.to_z3_uf(vmap, rset, wset)
    return {
      self.LT: e1 < e2,
      self.LE: e1 <= e2,
      self.GT: e1 > e2,
      self.GE: e1 >= e2,
      self.EQ: e1 == e2,
      self.NE: e1 != e2,
      }[self.op]

  def to_z3_nouf(self, vmap, tmap, rset, wset):
    e1 = self.e1.to_z3_nouf(vmap, tmap, rset, wset)
    e2 = self.e2.to_z3_nouf(vmap, tmap, rset, wset)
    return {
      self.LT: e1 < e2,
      self.LE: e1 <= e2,
      self.GT: e1 > e2,
      self.GE: e1 >= e2,
      self.EQ: e1 == e2,
      self.NE: e1 != e2,
      }[self.op]


################################
class Stmt(BaseClass):
  def __repr__(self):
    return self.tostr(0)

  def tostr(self, indent):
    raise Exception('tostr not implemented')

  def mk_indent(self, n):
    ret = ""
    for i in range(n):
      ret += "  "
    return ret


class Skip(Stmt):
  def tostr(self, indent):
    return self.mk_indent(indent) + "skip"

  def add_var_offset(self, off):
    return

  def collect_tmpls(self):
    return set()

  def collect_vars(self):
    return set()

def is_skip(p):
  return isinstance(p, Skip)


class If(Stmt):
  def __init__(self, bexp, s1):
    self.__init__(bexp, s1, Skip())

  def __init__(self, bexp, s1, s2):
    if is_skip(s1):
      self.bexp = bexp.neg()
      self.s1 = s2
      self.s2 = s1
    else:
      self.bexp = bexp
      self.s1 = s1
      self.s2 = s2
    assert(isinstance(self.bexp, BoolExpr))
    assert(isinstance(self.s1, Stmt))
    assert(isinstance(self.s2, Stmt))

  def tostr(self, indent):
    s  = self.mk_indent(indent) + "if (" + str(self.bexp) + ") {\n"
    s += self.s1.tostr(indent+1)
    s += "\n" + self.mk_indent(indent) + "}"
    if not isinstance(self.s2, Skip):
      s += " else {\n"
      s += self.s2.tostr(indent+1)
      s += "\n" + self.mk_indent(indent) + "}"
    return s

  def add_var_offset(self, off):
    self.bexp.add_var_offset(off)
    self.s1.add_var_offset(off)
    self.s2.add_var_offset(off)

  def collect_tmpls(self):
    s  = self.bexp.collect_tmpls()
    s |= self.s1.collect_tmpls()
    s |= self.s2.collect_tmpls()
    return s

  def collect_vars(self):
    s  = self.bexp.collect_vars()
    s |= self.s1.collect_vars()
    s |= self.s2.collect_vars()
    return s

def is_if(p):
  return isinstance(p, If)


class While(Stmt):
  def __init__(self, bexp, s1):
    self.__init__(bexp, s1, Skip())

  def __init__(self, bexp, body):
    self.bexp = bexp
    self.body = body
    assert(isinstance(self.bexp, BoolExpr))
    assert(isinstance(self.body, Stmt))

  def tostr(self, indent):
    s  = self.mk_indent(indent) + "while (" + str(self.bexp) + ") {\n"
    s += self.body.tostr(indent+1)
    s += "\n" + self.mk_indent(indent) + "}"
    return s

  def add_var_offset(self, off):
    self.bexp.add_var_offset(off)
    self.body.add_var_offset(off)

  def collect_tmpls(self):
    return self.bexp.collect_tmpls() | self.body.collect_tmpls()

  def collect_vars(self):
    return self.bexp.collect_vars() | self.body.collect_vars()

def is_while(p):
  return isinstance(p, While)


class StmtT(Stmt):
  def __init__(self, n):
    self.n = n
    assert(is_int(self.n))

  def tostr(self, indent):
    return self.mk_indent(indent) + "S" + str(self.n)

  def add_var_offset(self, off):
    return

  def collect_tmpls(self):
    return set([self])

  def collect_vars(self):
    return set()

  def to_z3_uf(self, vmap, rset, wset):
    rvars = [z3.If(Bool('R_%s_%s' % (self, v)), vmap[v], 0) for v in rset[self]]
    fn_sort = [IntSort() for i in range(len(rvars)+1)]

    for wv in wset[self]:
      fn = Function('%s_%s' % (self, wv), fn_sort)
      prefix = 'MW' if conf.use_must_write_fml else 'W'
      vmap[wv] = z3.If(Bool('%s_%s_%s' % (prefix, self, wv)), fn(rvars), vmap[wv])
    return BoolVal(True)

  def to_z3_nouf(self, vmap, tmap, rset, wset):
    rvars = [vmap[v] for v in rset[self]]

    for wv in wset[self]:
      var = FreshInt('%s_%s_' % (self, wv))
      if not tmap.has_key((self, wv)):
        tmap[(self, wv)] = []
      tmap[(self, wv)] += [(var, rvars)]
      prefix = 'MW' if conf.use_must_write_fml else 'W'
      vmap[wv] = z3.If(Bool('%s_%s_%s' % (prefix, self, wv)), var, vmap[wv])
    return BoolVal(True)

def is_stmtt(p):
  return isinstance(p, StmtT)


class Assign(Stmt):
  def __init__(self, lval, exp):
    self.lval = lval
    self.exp = exp
    assert(isinstance(self.lval, Lval))
    assert(isinstance(self.exp, Expr))

  def tostr(self, indent):
    s  = self.mk_indent(indent) + str(self.lval) + " := " + str(self.exp)
    return s

  def add_var_offset(self, off):
    self.lval.add_var_offset(off)
    self.exp.add_var_offset(off)

  def collect_tmpls(self):
    return self.exp.collect_tmpls()

  def collect_vars(self):
    return self.lval.collect_vars() | self.exp.collect_vars()

  def to_z3_uf(self, vmap, rset, wset):
    vmap[self.lval] = self.exp.to_z3_uf(vmap, rset, wset)
    return BoolVal(True)

  def to_z3_nouf(self, vmap, tmap, rset, wset):
    vmap[self.lval] = self.exp.to_z3_nouf(vmap, tmap, rset, wset)
    return BoolVal(True)

def is_assign(p):
  return isinstance(p, Assign)


class Compound(Stmt):
  def __init__(self, stmts):
    self.stmts = stmts
    assert(len(self.stmts) > 0)
    for s in self.stmts:
      assert(isinstance(s, Stmt))

  def tostr(self, indent):
    ret = ""
    first = True
    for s in self.stmts:
      if first:
        first = False
      else:
        ret += "\n"
      ret += s.tostr(indent)
    return ret

  def add_var_offset(self, off):
    for s in self.stmts:
      s.add_var_offset(off)

  def collect_tmpls(self):
    r = set()
    for s in self.stmts:
      r |= s.collect_tmpls()
    return r

  def collect_vars(self):
    r = set()
    for s in self.stmts:
      r |= s.collect_vars()
    return r

def is_compound(p):
  return isinstance(p, Compound)
