#!/usr/bin/env python

# PSyCO
# Copyright (C) 2014 Nuno Lopes
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import argparse, time, glob, re, conf
from z3 import *
from language import *
from parser import parse_str

Z3_SMTLIB_WP = 1
Z3_SMTLIB_WP_UNKNOWN = 2

global_msg_level = 0

def trace(n, msg):
  if global_msg_level & n != 0:
    print(msg)


# Global variables
WP         = BoolVal(True)  # weakest precondition found so far
extraWP    = BoolVal(True)  # mandatory WP conditions
MustWrite  = BoolVal(True)  # must write constraint (MW_x -> W_x)
MWVars     = []             # MW_x variables
wv_es_fml  = BoolVal(True)  # definition of DISJ_x_y variables
wv_es_vars = []             # AUX_* variables
cfg_encode_nouf = True      # whether to use UFs or not in SMT encoding


def mk_and(l):
  if len(l) == 0:
    return BoolVal(True)
  return And(l)


def inc(map, key, val):
  if map.has_key(key):
    map[key] += val
  else:
    map[key] = val


def print_table(t, name, suffix):
  if len(t) == 0:
    return

  print
  print name
  print '========'
  maxl = max([len(k) for k, v in t.iteritems()]) + 2
  for k, v in t.iteritems():
    print "%s:%s%s%s" % (k, ''.ljust(maxl-len(k)), v, suffix)


ticks = {}
def tick(n):
  inc(ticks, n, 1)

def print_ticks():
  print_table(ticks, 'Events', '')

watches = {}
def stop_watch(n, start):
  inc(watches, n, time.clock() - start)

def print_watches():
  print_table(watches, 'Time', 's')

def print_stats():
  print 'Precondition:'
  print WP
  print_ticks()
  print_watches()


def get_z3_id(x):
  return Z3_get_ast_id(x.ctx.ref(), x.as_ast())


def z3_solver_to_smtlib(s):
  a = s.assertions()
  size = len(a) - 1
  _a = (Ast * size)()
  for k in range(size):
    _a[k] = a[k].as_ast()

  return Z3_benchmark_to_smtlib_string(a[size].ctx_ref(), None, None, None, "",
                                       size, _a,  a[size].as_ast())


def gen_benchmark(s):
  pre = ("(set-info :source |\n Generated by PSyCO 0.1\n"
         " More info in N. P. Lopes and J. Monteiro. Weakest Precondition "
         "Synthesis for\n Compiler Optimizations, VMCAI'14.\n|)\n\n")
  string = pre + z3_solver_to_smtlib(s)

  files = glob.glob('bench/*.smt2')
  if len(files) == 0:
    return
  files.sort(reverse=True)
  filename = int(re.search('(\d+)\.smt2', files[0]).group(1)) + 1
  filename = 'bench/%03d.smt2' % filename
  fd = open(filename, 'w')
  fd.write(string)
  fd.close()


def to_z3(ls, vmap, tmap, rset, wset):
  if cfg_encode_nouf:
    f = [l.to_z3_nouf(vmap, tmap, rset, wset) for l in ls]
  else:
    f = [l.to_z3_uf(vmap, rset, wset) for l in ls]
  return simplify(mk_and(f))


def to_z3_nouf(ls, vmap, tmap, rset, wset):
  return simplify(mk_and([l.to_z3_nouf(vmap, tmap, rset, wset) for l in ls]))


def mk_var_equalities(m1, m2):
  return simplify(And([v == m2[k] for k, v in m1.iteritems()]))


# Cache of expressions generated by mk_uf_axioms()
# It's way faster to use Z3's substitute than build a new expression.
uf_exprs = {}

def get_uf_expr(tmpl, rset):
  if not uf_exprs.has_key(tmpl):
    e = And([Implies(Bool('R_%s_%s' % (tmpl, v)),
                    Int('%s_1' % v) == Int('%s_2' % v)) for v in rset[tmpl]])

    vars1 = [Int('%s_1' % v) for v in rset[tmpl]]
    vars2 = [Int('%s_2' % v) for v in rset[tmpl]]

    if is_boolexpt(tmpl):
      v1 = Bool('V_1')
      v2 = Bool('V_2')
    else:
      v1 = Int('V_1')
      v2 = Int('V_2')

    e = simplify(Implies(e, v1 == v2))
    uf_exprs[tmpl] = (e, [v1, v2] + vars1 + vars2)

  return uf_exprs[tmpl]


# apply the UF axiom:  x = y -> f(x) = f(y)
def mk_uf_axioms(tmap, rset):
  f = []
  for (tmpl, k), v in tmap.iteritems():
    e, tvars = get_uf_expr(tmpl, rset)
    for v1, vars1 in v:
      for v2, vars2 in v:
        if get_z3_id(v1) >= get_z3_id(v2):
          continue
        f += [substitute(e, zip(tvars, [v1,v2] + vars1 + vars2))]
  return mk_and(f)


def mk_uf_axioms2(tmap1, tmap2, rset):
  f = []
  for (tmpl, k), vs1 in tmap1.iteritems():
    if not tmap2.has_key((tmpl, k)):
      continue
    vs2 = tmap2[(tmpl, k)]
    e, tvars = get_uf_expr(tmpl, rset)
    for v1, vars1 in vs1:
      for v2, vars2 in vs2:
        f += [substitute(e, zip(tvars, [v1,v2] + vars1 + vars2))]
  return f


def append_paths(ps1, ps2):
  ps = []
  for p1 in ps1:
    for p2 in ps2:
      ps += [p1 + p2]
  return ps


def join_paths(ps1, ps2):
  copy = list(ps1)
  copy.extend(ps2)
  return copy


def gen_paths_(p, paths, unroll):
  if is_skip(p):
    return paths

  if is_assign(p) or is_stmtt(p):
    return append_paths(paths, [[p]])

  if is_compound(p):
    for s in p.stmts:
      paths = gen_paths_(s, paths, unroll)
    return paths

  if is_if(p):
    pos = gen_paths_(p.s1, append_paths(paths, [[p.bexp]]), unroll)
    neg = gen_paths_(p.s2, append_paths(paths, [[p.bexp.neg()]]), unroll)
    return join_paths(pos, neg)

  if is_while(p):
    negbexp = [[p.bexp.neg()]]
    ps = append_paths(paths, negbexp)
    last_loop = paths
    for i in range(unroll):
      last_loop = gen_paths_(p.body, append_paths(last_loop, [[p.bexp]]), unroll)
      ps = join_paths(ps, append_paths(last_loop, negbexp))
    return ps

  raise Exception('unknown stmt')


def gen_paths(p, unroll):
  return gen_paths_(p, [[]], unroll)


def create_path_fmls(p, vmap0, rset, wset, unroll):
  paths = gen_paths(p, unroll)
  fmls = []
  for p in paths:
    vmap = dict(vmap0)
    tmap = {}
    f = to_z3(p, vmap, tmap, rset, wset)
    f = simplify(And(f, mk_uf_axioms(tmap, rset)))
    fmls += [(p, f, vmap, tmap)]
  return fmls


def slow_fml_simplify(fml):
  # TODO: benchmark: simplify(fml, pull_cheap_ite=True, ite_extra_rules=True,
  #                                push_ite_arith=True, local_ctx=True)
  t1 = Tactic('simplify')
  t2 = Tactic('ctx-simplify')
  t3 = Tactic('propagate-values')
  t4 = Tactic('elim-term-ite')
  # t = Then(t1, t2, t3, t4, t1)
  t = Then(t1, t2, t3, t1)
  return t.apply(fml)[0].as_expr()


def extract_implied_RW(fml, rset, wset):
  s = Solver()
  s.add(wv_es_fml)
  s.add(fml)
  assert s.check() == sat

  vars = [Not(v) for v in wv_es_vars]
  for tmpl, vs in rset.iteritems():
    vars += [Bool('R_%s_%s' % (tmpl, v)) for v in vs]

  for tmpl, vs in wset.iteritems():
    vars += [Bool('W_%s_%s' % (tmpl, v)) for v in vs]

  return mk_and([Not(v) for v in vars if s.check(v) == unsat])


def z3_simplify_wp(fml, rset, wset):
  implied = extract_implied_RW(fml, rset, wset)
  fml = And(implied, fml)

  t1 = Tactic('simplify')
  t2 = Tactic('ctx-solver-simplify')
  t3 = Tactic('propagate-values')
  t = Then(t3, t1, t2, t1, t2)
  return t.apply(fml)[0].as_expr()


# linear, left-to-right, core minimization
def minimize_core(s, vars):
  core = []
  for i in range(len(vars)):
    f = core + vars[i+1:]
    if s.check(f) != unsat:
      core += [vars[i]]

  if len(core) < len(vars):
    tick('core_not_minimal')

  return And(core)


def generalize_model_(s, BiasVs, PosVs, NegVs):
  BiasVs.sort()

  # if bias vars are enough, use them since they produce shorter WPs
  if s.check(BiasVs) == unsat:
    return minimize_core(s, list(s.unsat_core())), True

  # if negative vars are enough, skip analysis of positive vars
  if s.check(NegVs) == unsat:
    return minimize_core(s, list(s.unsat_core()) + BiasVs), True

  tick('num_pos_wps')
  sol = s.check(NegVs + PosVs)
  assert sol == unsat
  return minimize_core(s, list(s.unsat_core()) + BiasVs), False


def generalize_model(s, BiasVs, PosVs, NegVs):
  start = time.clock()
  r = generalize_model_(s, BiasVs, PosVs, NegVs)
  stop_watch('generalize_model', start)
  return r


def extract_WP(s, s2, rset, wset):
  global WP

  NewWP = []
  while True:
    start = time.clock()
    gen_benchmark(s)
    status = s.check()
    stop_watch('solve_wp', start)

    if status != sat:
      break

    tick('num_models')
    start = time.clock()
    model = s.model()
    ModelPosVars = []
    ModelNegVars = []
    ModelBiasVars = []

    for var in model.decls():
      varname = str(var)
      bias = False
      if varname.startswith('AUX_'):
        bias = True
      elif not (varname.startswith('R_') or varname.startswith('W_')):
        continue

      tick('num_model_vars')
      if is_true(model[var]):
        if bias:
          ModelBiasVars += [Bool(varname)]
        else:
          ModelPosVars += [Bool(varname)]
      elif not bias:
        ModelNegVars += [Not(Bool(varname))]
    stop_watch('read_wp_model', start)

    NewWPSol, OK = generalize_model(s2, ModelBiasVars, ModelPosVars,
                                    ModelNegVars)
    s.add(Not(NewWPSol))
    if OK:
      NewWP += [z3_simplify_wp(NewWPSol, rset, wset)]

  if status != unsat:
    trace(Z3_SMTLIB_WP_UNKNOWN, z3_solver_to_smtlib(s))
    tick('num_z3_unknown')
    print "Z3 returned unknown: weakest precondition not guaranteed"

  if NewWP == []:
    print "refine_wp: no generalized model avaliable (TODO!)"
    exit()

  start = time.clock()
  WP = z3_simplify_wp(And(WP, Or(NewWP)), rset, wset)
  stop_watch('simplify_wp', start)


def refine_wp_(p1, p2, vmap01, vmap02, rset, wset):
  global WP

  s = Solver()
  #s = SimpleSolver()
  #s.set('auto_config', False)
  #s.set('smt.phase_selection', 1)  # assign true to literals first
  #s.set('model.compact', True)
  #s.set('model.partial', True)

  start = time.clock()

  tmap = {}
  vmap1 = dict(vmap01)
  p1 = to_z3_nouf(p1, vmap1, tmap, rset, wset)
  vmap2 = dict(vmap02)
  p2 = to_z3_nouf(p2, vmap2, tmap, rset, wset)

  path      = simplify(And(p1, p2))
  uf_axioms = mk_uf_axioms(tmap, rset)
  var_eqs   = mk_var_equalities(vmap1, vmap2)

  # gather universally quantified variables
  univ_vars  = [v for v in vmap01.itervalues()]
  univ_vars += [v for k, v in vmap02.iteritems() if not vmap01.has_key(k)]
  univ_vars += MWVars
  for k, v in tmap.iteritems():
    univ_vars += [ufv for ufv,in_vs in v]

  f = And(WP, extraWP, ForAll(univ_vars, Implies(And(uf_axioms, MustWrite),
                                                 Or(Not(path), var_eqs))))
  f = slow_fml_simplify(f)
  s.add(f)

  trace(Z3_SMTLIB_WP, z3_solver_to_smtlib(s))

  # new Solver with the negated formula for model generalization
  s2 = Solver()
  s2.add(slow_fml_simplify(And(extraWP, WP, uf_axioms, MustWrite, path,
                               Not(var_eqs))))
  assert s2.check() == sat
  stop_watch('create_wp_fml', start)

  extract_WP(s, s2, rset, wset)


def refine_wp(p1, p2, vmap01, vmap02, rset, wset):
  start = time.clock()
  refine_wp_(p1, p2, vmap01, vmap02, rset, wset)
  stop_watch('refine_wp', start)


def check_equiv(s, itern, paths1, paths2, skipn, rset, wset, vmap01, vmap02,
                unroll):
  start = time.clock()
  s.add(WP)

  print('------------------------------\n')
  print('Checking paths w/ unroll=%d skip=%d iteration=%d' %
        (unroll, skipn, itern))

  i = 0
  for path1, f1, vmap1, tmap1 in paths1:
    if i + len(paths2) <= skipn:
      i += len(paths2)
      continue

    s.push()
    s.add(f1)

    for path2, f2, vmap2, tmap2 in paths2:
      i += 1
      if i <= skipn:
        continue

      s.push()
      s.add(f2)

      if s.check() == unsat:
        s.pop()
        continue

      t2 = time.clock()
      s.add(mk_uf_axioms2(tmap1, tmap2, rset))
      s.add(Not(mk_var_equalities(vmap1, vmap2)))
      stop_watch('fp_fml_gen', t2)

      if s.check() != unsat:
        s.pop(2)
        stop_watch('fixedpoint', start)
        print('\nnon-equivalent path:\n%s\n%s' % (path1, path2))
        refine_wp(path1, path2, vmap01, vmap02, rset, wset)
        return i
      s.pop()
    s.pop()
  stop_watch('fixedpoint', start)
  return i


# Create AUX_* vars
def create_aux_vars(rset, wset, p1vars, vars_and_ctx):
  global extraWP, wv_es_fml, wv_es_vars

  wv_es_fml  = []
  wv_es_vars = []
  cross      = []
  cross3     = []

  rs_vars = ['R_%s' % tmpl for tmpl in rset.iterkeys()]
  ws_vars = ['W_%s' % tmpl for tmpl, vs in wset.iteritems() if len(vs) > 0]

  for wv in ws_vars:
    cross += [(wv, rv) for rv in rs_vars]
    cross += [(wv, wv2) for wv2 in ws_vars if wv < wv2]

  for wv in ws_vars:
    for wv2 in ws_vars:
      if wv >= wv2:
        continue
      cross3 += [(wv, wv2, rv) for rv in rs_vars]
      #cross3 += [(wv, wv2, wv3) for wv3 in ws_vars if wv2 < wv3]

  # DISJ_x_y vars
  for v1, v2 in cross:
    vs = []
    for v in vars_and_ctx:
      vs += [And(Bool('%s_%s' % (v1, v)), Bool('%s_%s' % (v2, v)))]
    var = Bool('AUX_DISJ_%s_%s' % (v1, v2))
    wv_es_vars += [var]
    wv_es_fml  += [var == Not(Or(vs))]

  # DISJ3_x_y vars
  for v1, v2, v3 in cross3:
    vs = []
    for v in vars_and_ctx:
      vs += [And(Bool('%s_%s' % (v1, v)), Bool('%s_%s' % (v2, v)),
                 Bool('%s_%s' % (v3, v)))]
    var = Bool('AUX_DISJ3_%s_%s_%s' % (v1, v2, v3))
    wv_es_vars += [var]
    wv_es_fml  += [var == Not(Or(vs))]

  # INTR_x_y_z vars
  for v1, v2 in cross:
    for v3 in rs_vars + ws_vars:
      #if v3 == v1 or v3 == v2:
      #  continue
      vs = []
      for v in vars_and_ctx:
        vs += [And(Bool('%s_%s' % (v1, v)), Bool('%s_%s' % (v2, v))) ==
               Bool('%s_%s' % (v3, v))]
      var = Bool('AUX_INTR_%s_%s_%s' % (v1,v2,v3))
      wv_es_vars += [var]
      wv_es_fml  += [var == And(vs)]

  # EQ_x_y_z vars
  for v1, v2 in cross:
    for eqv in p1vars:
      vs = []
      for v in vars_and_ctx:
        if v == eqv:
          vs += [Bool('%s_%s' % (v1, v)), Bool('%s_%s' % (v2, v))]
        else:
          vs += [Not(And(Bool('%s_%s' % (v1, v)), Bool('%s_%s' % (v2, v))))]
      var = Bool('AUX_EQ_%s_%s_%s' % (v1,v2,eqv))
      wv_es_vars += [var]
      wv_es_fml  += [var == And(vs)]

  # EQ2_w_x_y_z vars
  for v1, v2 in cross:
    for eqv in p1vars:
      for eqv2 in p1vars:
        if eqv >= eqv2:
          continue
        vs = []
        for v in vars_and_ctx:
          if v == eqv or v == eqv2:
            vs += [Bool('%s_%s' % (v1, v)), Bool('%s_%s' % (v2, v))]
          else:
            vs += [Not(And(Bool('%s_%s' % (v1, v)), Bool('%s_%s' % (v2, v))))]
        var = Bool('AUX_EQ2_%s_%s_%s_%s' % (v1,v2,eqv,eqv2))
        wv_es_vars += [var]
        wv_es_fml  += [var == And(vs)]

  wv_es_fml = mk_and(wv_es_fml)
  extraWP = Tactic('propagate-values')(And(extraWP, wv_es_fml))[0].as_expr()


def main():
  parser = argparse.ArgumentParser(description='PSyCO 0.1')
  parser.add_argument('--file', dest='file', required=True, type=str,
                      help='Prefix of optimization filename to analyze.')
  parser.add_argument('--unroll', dest='unroll', required=False, type=int,
                      default=3, help='Number of times to unroll each loop.')
  parser.add_argument('--use-mw', dest='use_mw', required=False, default=True,
                      action='store_true', help='Use must-write constraints.')
  parser.add_argument('--no-use-mw', dest='use_mw', required=False,
                      default=True, action='store_false',
                      help='Don\'t use must-write constraints.')
  parser.add_argument('--use-uf', dest='use_uf', required=False,
                      action='store_false', help='Use UFs in constraints.')
  parser.add_argument('--wp-dnf', dest='wp_dnf', required=False, default=False,
                      action='store_true', help='Output WP in DNF.')
  parser.add_argument('--verbose', dest='verbose', required=False, type=int,
                      default=0, help='Bitmask to enable debug information.')

  start = time.clock()

  args = parser.parse_args()

  conf.init()

  global global_msg_level, cfg_encode_nouf
  global_msg_level = args.verbose
  conf.use_must_write_fml = args.use_mw
  cfg_encode_nouf = args.use_uf

  p1 = open(args.file + '.orig', 'r')
  p1 = parse_str(p1.read())

  p2 = open(args.file + '.opt', 'r')
  p2 = parse_str(p2.read())

  stop_watch('parsing', start)

  s2 = time.clock()
  rset = {}
  wset = {}
  tmpls = p1.collect_tmpls() | p2.collect_tmpls()

  stmts = [s for s in tmpls if is_stmtt(s)]
  p1vars = p1.collect_vars()
  p2vars = p2.collect_vars()

  all_vars = p1vars.union(p2vars)
  max_var = max([-1] + [v.n for v in all_vars])

  # 1 <= context vars <= num_stmts + 1
  num_ctx_vars = len(stmts) + 1
  vars_and_ctx = set(p1vars)
  vars_and_ctx |= set([VarL(i) for i in range(max_var+1,
                                              max_var + num_ctx_vars + 1)])

  # use all vars as read/write sets by default
  for t in tmpls:
    rset[t] = vars_and_ctx
    wset[t] = vars_and_ctx if is_stmtt(t) else []

  global extraWP, wv_es_fml, wv_es_vars, MustWrite, MWVars

  # S_i always writes to C_i.
  extraWP = [Bool('W_%s_%s' % (s, VarL(max_var + s.n))) for s in stmts]
  extraWP = simplify(mk_and(extraWP))

  create_aux_vars(rset, wset, p1vars, vars_and_ctx)

  # create the initial set of vars (shared by both progs)
  vmap01 = {}
  for v in vars_and_ctx:
    vmap01[v] = Int('%s_0' % v)

  vmap02 = dict(vmap01)
  for v in p2vars:
    vmap02[v] = Int('%s_0' % v)

  # generate path constraints
  paths1 = create_path_fmls(p1, vmap01, rset, wset, args.unroll)
  paths2 = create_path_fmls(p2, vmap02, rset, wset, args.unroll)
  num_paths = len(paths1) * len(paths2)


  # generate the following constraints:  must-write -> write
  MustWrite = []
  MWVars = []
  if conf.use_must_write_fml:
    for stmt, vs in wset.iteritems():
      vs = ['W_%s_%s' % (stmt, v) for v in vs]
      for v in vs:
        mw = Bool('M%s' % v)
        w = Bool(v)
        MustWrite += [Implies(mw, w)]
        MWVars += [mw]
  MustWrite = simplify(mk_and(MustWrite))

  stop_watch('gen_paths', s2)


  print('Program 1\n=========\n%s\n' % p1)
  print('Program 2\n=========\n%s\n' % p2)

  print('\n------------------------------')
  print('%d paths in P1 and %d paths in P2  (cross = %d)\n' %
        (len(paths1), len(paths2), num_paths))


  s = Solver()
  s.add(extraWP)
  s.add(MustWrite)

  itern = 0
  checked = 0

  while True:
    checked = check_equiv(s, itern, paths1, paths2, checked, rset, wset, vmap01,
                          vmap02, args.unroll)
    if checked >= num_paths:
      break
    itern += 1

  # Put WP in DNF form if requested
  if args.wp_dnf:
    global WP
    s = Solver()
    s.add(extraWP)
    s.add(WP)
    assert s.check() == sat

    s2 = Solver()
    s2.add(wv_es_fml)
    s2.add(Not(WP))

    if s2.check() == sat:
      extract_WP(s, s2, rset, wset)

  stop_watch('total', start)

  print "Done!\n"
  print_stats()


if __name__ == "__main__":
  try:
    main()
  except KeyboardInterrupt:
    print '\nCaught Ctrl-C. Exiting..'
    print_stats()
