# PSyCO
# Copyright (C) 2014 Nuno Lopes
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from lrparsing import *
from language import *

class Parser(Grammar):
  class T(TokenRegistry):
    integer = Token(re="[0-9]+")
    newline = Token(re="\r\n?|\n")

  lval = Token('V') + T.integer

  num = T.integer

  expr = Ref('expr')
  expr_p = Token('(') + expr + Token(')')
  expt = Token('E') + num
  min = Token('min') + Token('(') + expr + Token(',') + expr + Token(')')
  max = Token('max') + Token('(') + expr + Token(',') + expr + Token(')')

  expr = Prio(
    num,
    lval,
    expt,
    Tokens("+ -") >> THIS,
    THIS << Tokens("* /") << THIS,
    THIS << Tokens("+ -") << THIS,
    expr_p,
    min,
    max,
  )

  boolexpr = Ref('boolexpr')
  true = Token('true')
  false = Token('false')
  bexpt = Token('B') + num
  boolexpr_p = Token('(') + boolexpr + Token(')')

  boolexpr = Prio(
    Token('!') << THIS,
    THIS << Token('&&') << THIS,
    THIS << Token('||') << THIS,
    expr << Tokens("= != < <= > >=") << expr,
    true,
    false,
    bexpt,
    boolexpr_p,
  )

  stmt = Ref('stmt')
  stmts = Ref('stmts')

  stmtt = Token('S') + T.integer
  assgn = lval + Token(':=') + expr
  skip = Token('skip')
  then = Token('{') + T.newline + stmts + Opt(T.newline) + Token('}')
  elses = Opt(Token('else') + then)
  ifs = Token('if') + Token('(') + boolexpr + Token(')') + then + elses
  whiles = Token('while') + Token('(') + boolexpr + Token(')') + then

  stmt = Prio(
    stmtt,
    ifs,
    whiles,
    assgn,
    skip,
  )


  stmts2 = stmts + T.newline + stmt

  stmts = Prio(
    stmt,
    THIS + T.newline,
    stmts2
  )

  START = stmts

  COMMENTS = (
    Token(re="//(?:[^\r\n]*(?:\r\n?|\n))") |
    Token(re="/[*](?:[^*]|[*][^/])*[*]/")
  )

  ### AST construction
  @classmethod
  def lang_factory(cls, term):
    return term[1] if not "eval" in term[0] else term[0]["eval"](term)

  def eval_bexpr(term):
    # NOT expr
    if len(term) == 3:
      return term[2].neg()

    # true/false
    if len(term) == 2:
      return term[1]

    assert(len(term) == 4)
    a = term[1]
    b = term[3]

    if term[2] == '&&':
      return AndExpr(a, b)
    if term[2] == '||':
      return OrExpr(a, b)
    return {
      '=':  CmpOp(a, b, CmpOp.EQ),
      '!=': CmpOp(a, b, CmpOp.NE),
      '>':  CmpOp(a, b, CmpOp.GT),
      '>=': CmpOp(a, b, CmpOp.GE),
      '<':  CmpOp(a, b, CmpOp.LT),
      '<=': CmpOp(a, b, CmpOp.LE),
      }[term[2]]

  def eval_expr(term):
    # number
    if len(term) == 2:
      return term[1]

    # unary op
    if len(term) == 3:
      a = term[2]
      if term[1] == '+':
        return a
      elif is_num(a):
        a.n = -a.n
        return a
      else:
        return BinOp(Num(-1), a, BinOp.Mul)

    assert(len(term) == 4)
    a = term[1]
    b = term[3]
    return {
      '+': BinOp(a, b, BinOp.Add),
      '-': BinOp(a, b, BinOp.Sub),
      '*': BinOp(a, b, BinOp.Mul),
      '/': BinOp(a, b, BinOp.Div),
      }[term[2]]

  boolexpr["eval"]   = eval_bexpr
  boolexpr_p["eval"] = lambda n: n[2]
  true["eval"] =       lambda n: TrueExpr()
  false["eval"] =      lambda n: FalseExpr()
  bexpt["eval"] =      lambda n: BoolExpT(n[2].n)
  num["eval"] =        lambda n: Num(int(n[1]))
  lval["eval"] =       lambda n: VarL(n[2].n)
  expr["eval"] =       eval_expr
  expr_p["eval"] =     lambda n: n[2]
  expt["eval"] =       lambda n: ExpT(n[2].n)
  min["eval"] =        lambda n: Min(n[3], n[5])
  max["eval"] =        lambda n: Max(n[3], n[5])
  stmtt["eval"] =      lambda n: StmtT(n[2].n)
  assgn["eval"] =      lambda n: Assign(n[1], n[3])
  skip["eval"] =       lambda n: Skip()
  then["eval"] =       lambda n: n[3]
  elses["eval"] =      lambda n: n[2] if len(n) > 1 else Skip()
  ifs["eval"] =        lambda n: If(n[3], n[5], n[6])
  whiles["eval"] =     lambda n: While(n[3], n[5])
  stmts2["eval"] =     lambda n: Compound([n[1], n[3]])


def parse_str(txt):
  try:
    if Parser.pre_compile_grammar(Parser.pre_compiled) != None:
      raise Exception('outdated parser')
    return Parser.parse(txt, Parser.lang_factory)
  except LrParsingError, e:
    print(e)
    exit(-1)
  #print(Parser.repr_parse_tree(Parser.parse(txt)))


Parser.pre_compiled = ('2858c73ec9a1773658283c9e3ba3828c4b72db52b5ebeb81fab3df16a458a89918cdac5474d1d52815f77bac4e0af740e7b8a89034f20dc12a9845c34097e3f4', 0, ({"'S'": 20, "'skip'": 24, "'if'": 21, "'while'": 22, "'V'": 25}, {32: 23, 2: 1, 4: 2, 5: 3, 6: 4, 7: 5, 8: 6, 9: 7, 11: 8, 12: 9, 13: 10, 14: 11, 15: 12, 16: 13, 17: 14, 18: 15, 19: 16, 20: 17, 21: 18, 22: 19}, ['<Parser>']), ({'__end_of_input__': 26}, {}, ['<Parser>']), ({'__end_of_input__': (2, 1, 'START'), 'T.newline': 27}, {}, ['START', 'stmts', 'stmts2']), ({'__end_of_input__': (4, 1, 'stmts'), "'}'": (4, 1, 'stmts'), 'T.newline': (4, 1, 'stmts')}, {}, ['stmts']), ({'__end_of_input__': (5, 1), "'}'": (5, 1), 'T.newline': (5, 1)}, {}, ['stmts']), ({'__end_of_input__': (5, 1), "'}'": (5, 1), 'T.newline': (5, 1)}, {}, ['stmts']), ({'__end_of_input__': (5, 1), "'}'": (5, 1), 'T.newline': (5, 1)}, {}, ['stmts']), ({'__end_of_input__': (6, 1), "'}'": (6, 1), 'T.newline': (6, 1)}, {}, ['stmts']), ({'__end_of_input__': (8, 1), "'}'": (8, 1), 'T.newline': (8, 1)}, {}, ['stmts']), ({'__end_of_input__': (9, 1, 'stmt'), "'}'": (9, 1, 'stmt'), 'T.newline': (9, 1, 'stmt')}, {}, ['stmt']), ({'__end_of_input__': (12, 1), "'}'": (12, 1), 'T.newline': (12, 1)}, {}, ['stmt']), ({'__end_of_input__': (12, 1), "'}'": (12, 1), 'T.newline': (12, 1)}, {}, ['stmt']), ({'__end_of_input__': (12, 1), "'}'": (12, 1), 'T.newline': (12, 1)}, {}, ['stmt']), ({'__end_of_input__': (12, 1), "'}'": (12, 1), 'T.newline': (12, 1)}, {}, ['stmt']), ({'__end_of_input__': (12, 1), "'}'": (12, 1), 'T.newline': (12, 1)}, {}, ['stmt']), ({'__end_of_input__': (13, 1), "'}'": (13, 1), 'T.newline': (13, 1)}, {}, ['stmt']), ({'__end_of_input__': (14, 1), "'}'": (14, 1), 'T.newline': (14, 1)}, {}, ['stmt']), ({'__end_of_input__': (15, 1), "'}'": (15, 1), 'T.newline': (15, 1)}, {}, ['stmt']), ({'__end_of_input__': (16, 1), "'}'": (16, 1), 'T.newline': (16, 1)}, {}, ['stmt']), ({'__end_of_input__': (17, 1), "'}'": (17, 1), 'T.newline': (17, 1)}, {}, ['stmt']), ({'T.integer': 29}, {24: 28}, ['stmtt']), ({"'('": 30}, {}, ['ifs']), ({"'('": 31}, {}, ['whiles']), ({"':='": 32}, {}, ['assgn']), ({'__end_of_input__': (22, 1, 'skip'), "'}'": (22, 1, 'skip'), 'T.newline': (22, 1, 'skip')}, {}, ['skip']), ({'T.integer': 254}, {24: 158}, ['lval']), ({'__empty__': (0, 2, '<Parser>')}, {}, ['<Parser>']), ({"'skip'": 24, 'T.newline': (7, 2), '__end_of_input__': (7, 2), "'V'": 25, "'if'": 21, "'S'": 20, "'while'": 22}, {32: 23, 9: 34, 12: 9, 13: 10, 14: 11, 15: 12, 16: 13, 17: 14, 18: 15, 19: 16, 20: 17, 21: 18, 22: 19}, ['stmts', 'stmts2']), ({'__end_of_input__': (18, 2, 'stmtt'), "'}'": (18, 2, 'stmtt'), 'T.newline': (18, 2, 'stmtt')}, {}, ['stmtt']), ({"'!='": (24, 1, 'num'), "'>'": (24, 1, 'num'), 'T.newline': (24, 1, 'num'), '__end_of_input__': (24, 1, 'num'), "'='": (24, 1, 'num'), "'/'": (24, 1, 'num'), "'-'": (24, 1, 'num'), "','": (24, 1, 'num'), "')'": (24, 1, 'num'), "'<='": (24, 1, 'num'), "':='": (24, 1, 'num'), "'+'": (24, 1, 'num'), "'||'": (24, 1, 'num'), "'>='": (24, 1, 'num'), "'*'": (24, 1, 'num'), "'}'": (24, 1, 'num'), "'&&'": (24, 1, 'num'), "'<'": (24, 1, 'num')}, {}, ['num']), ({"'B'": 85, "'true'": 83, "'min'": 60, "'false'": 84, 'T.integer': 29, "'max'": 61, "'-'": 59, "'V'": 25, "'('": 62, "'E'": 57, "'!'": 82, "'+'": 58}, {24: 35, 27: 159, 32: 37, 34: 64, 37: 65, 42: 40, 43: 66, 44: 67, 45: 68, 46: 69, 47: 70, 48: 71, 49: 72, 50: 73, 51: 41, 52: 42, 53: 43, 54: 44, 55: 45, 56: 46, 57: 47, 58: 48, 59: 49, 60: 74, 61: 75, 62: 76, 63: 77, 64: 78, 65: 79, 66: 80, 67: 81, 68: 50, 69: 51, 70: 52, 71: 53, 72: 54, 73: 55, 74: 56}, ['ifs']), ({"'B'": 85, "'true'": 83, "'min'": 60, "'false'": 84, 'T.integer': 29, "'max'": 61, "'-'": 59, "'V'": 25, "'('": 62, "'E'": 57, "'!'": 82, "'+'": 58}, {24: 35, 27: 63, 32: 37, 34: 64, 37: 65, 42: 40, 43: 66, 44: 67, 45: 68, 46: 69, 47: 70, 48: 71, 49: 72, 50: 73, 51: 41, 52: 42, 53: 43, 54: 44, 55: 45, 56: 46, 57: 47, 58: 48, 59: 49, 60: 74, 61: 75, 62: 76, 63: 77, 64: 78, 65: 79, 66: 80, 67: 81, 68: 50, 69: 51, 70: 52, 71: 53, 72: 54, 73: 55, 74: 56}, ['whiles']), ({"'min'": 60, 'T.integer': 39, "'max'": 61, "'-'": 59, "'V'": 25, "'('": 36, "'E'": 57, "'+'": 58}, {32: 37, 34: 38, 68: 50, 69: 51, 70: 52, 71: 53, 72: 54, 73: 55, 42: 40, 56: 46, 51: 41, 52: 42, 53: 43, 54: 44, 55: 45, 24: 35, 57: 47, 58: 48, 59: 49, 74: 56}, ['assgn']), ({"'!='": (32, 2, 'lval'), "'>'": (32, 2, 'lval'), 'T.newline': (32, 2, 'lval'), '__end_of_input__': (32, 2, 'lval'), "'/'": (32, 2, 'lval'), "'-'": (32, 2, 'lval'), "','": (32, 2, 'lval'), "')'": (32, 2, 'lval'), "'<='": (32, 2, 'lval'), "':='": (32, 2, 'lval'), "'+'": (32, 2, 'lval'), "'||'": (32, 2, 'lval'), "'>='": (32, 2, 'lval'), "'*'": (32, 2, 'lval'), "'='": (32, 2, 'lval'), "'&&'": (32, 2, 'lval'), "'<'": (32, 2, 'lval')}, {}, ['lval']), ({'__end_of_input__': (11, 3, 'stmts2'), "'}'": (11, 3, 'stmts2'), 'T.newline': (11, 3, 'stmts2')}, {}, ['stmts2']), ({"'!='": (51, 1), "'>'": (51, 1), 'T.newline': (51, 1), '__end_of_input__': (51, 1), "'='": (51, 1), "'/'": (51, 1), "'-'": (51, 1), "','": (51, 1), "')'": (51, 1), "'<='": (51, 1), "'>='": (51, 1), "'*'": (51, 1), "'}'": (51, 1), "'+'": (51, 1), "'<'": (51, 1)}, {}, ['expr']), ({"'min'": 60, 'T.integer': 39, "'max'": 61, "'-'": 59, "'V'": 25, "'('": 36, "'E'": 57, "'+'": 58}, {32: 37, 34: 114, 68: 50, 69: 51, 70: 52, 71: 53, 72: 54, 73: 55, 42: 40, 56: 46, 51: 41, 52: 42, 53: 43, 54: 44, 55: 45, 24: 35, 57: 47, 58: 48, 59: 49, 74: 56}, ['expr_p']), ({"'!='": (52, 1), "'>'": (52, 1), 'T.newline': (52, 1), '__end_of_input__': (52, 1), "'='": (52, 1), "'/'": (52, 1), "'-'": (52, 1), "','": (52, 1), "')'": (52, 1), "'<='": (52, 1), "'>='": (52, 1), "'*'": (52, 1), "'}'": (52, 1), "'+'": (52, 1), "'<'": (52, 1)}, {}, ['expr']), ({'T.newline': (21, 3, 'assgn'), '__end_of_input__': (21, 3, 'assgn'), "'/'": 154, "'-'": 152, "'*'": 153, "'}'": (21, 3, 'assgn'), "'+'": 151}, {}, ['assgn', 'expr']), ({"'!='": (24, 1, 'num'), "'>'": (24, 1, 'num'), 'T.newline': (24, 1, 'num'), '__end_of_input__': (24, 1, 'num'), "'='": (24, 1, 'num'), "'/'": (24, 1, 'num'), "'-'": (24, 1, 'num'), "','": (24, 1, 'num'), "')'": (24, 1, 'num'), "'<='": (24, 1, 'num'), "':='": (24, 1, 'num'), "'+'": (24, 1, 'num'), "'||'": (24, 1, 'num'), "'>='": (24, 1, 'num'), "'*'": (24, 1, 'num'), "'}'": (24, 1, 'num'), "'&&'": (24, 1, 'num'), "'<'": (24, 1, 'num')}, {}, ['num']), ({"'!='": (34, 1, 'expr'), "'>'": (34, 1, 'expr'), 'T.newline': (34, 1, 'expr'), '__end_of_input__': (34, 1, 'expr'), "'='": (34, 1, 'expr'), "'/'": (34, 1, 'expr'), "'-'": (34, 1, 'expr'), "','": (34, 1, 'expr'), "')'": (34, 1, 'expr'), "'<='": (34, 1, 'expr'), "'>='": (34, 1, 'expr'), "'*'": (34, 1, 'expr'), "'}'": (34, 1, 'expr'), "'+'": (34, 1, 'expr'), "'<'": (34, 1, 'expr')}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (53, 1), "'>'": (53, 1), 'T.newline': (53, 1), '__end_of_input__': (53, 1), "'='": (53, 1), "'/'": (53, 1), "'-'": (53, 1), "','": (53, 1), "')'": (53, 1), "'<='": (53, 1), "'>='": (53, 1), "'*'": (53, 1), "'}'": (53, 1), "'+'": (53, 1), "'<'": (53, 1)}, {}, ['expr']), ({"'!='": (54, 1), "'>'": (54, 1), 'T.newline': (54, 1), '__end_of_input__': (54, 1), "'='": (54, 1), "'/'": (54, 1), "'-'": (54, 1), "','": (54, 1), "')'": (54, 1), "'<='": (54, 1), "'>='": (54, 1), "'*'": (54, 1), "'}'": (54, 1), "'+'": (54, 1), "'<'": (54, 1)}, {}, ['expr']), ({"'!='": (55, 1), "'>'": (55, 1), 'T.newline': (55, 1), '__end_of_input__': (55, 1), "'='": (55, 1), "'/'": (55, 1), "'-'": (55, 1), "','": (55, 1), "')'": (55, 1), "'<='": (55, 1), "'>='": (55, 1), "'*'": (55, 1), "'}'": (55, 1), "'+'": (55, 1), "'<'": (55, 1)}, {}, ['expr']), ({"'!='": (56, 1), "'>'": (56, 1), 'T.newline': (56, 1), '__end_of_input__': (56, 1), "'='": (56, 1), "'/'": (56, 1), "'-'": (56, 1), "','": (56, 1), "')'": (56, 1), "'<='": (56, 1), "'>='": (56, 1), "'*'": (56, 1), "'}'": (56, 1), "'+'": (56, 1), "'<'": (56, 1)}, {}, ['expr']), ({"'!='": (57, 1), "'>'": (57, 1), 'T.newline': (57, 1), '__end_of_input__': (57, 1), "'='": (57, 1), "'/'": (57, 1), "'-'": (57, 1), "','": (57, 1), "')'": (57, 1), "'<='": (57, 1), "'>='": (57, 1), "'*'": (57, 1), "'}'": (57, 1), "'+'": (57, 1), "'<'": (57, 1)}, {}, ['expr']), ({"'!='": (58, 1), "'>'": (58, 1), 'T.newline': (58, 1), '__end_of_input__': (58, 1), "'='": (58, 1), "'/'": (58, 1), "'-'": (58, 1), "','": (58, 1), "')'": (58, 1), "'<='": (58, 1), "'>='": (58, 1), "'*'": (58, 1), "'}'": (58, 1), "'+'": (58, 1), "'<'": (58, 1)}, {}, ['expr']), ({"'!='": (59, 1), "'>'": (59, 1), 'T.newline': (59, 1), '__end_of_input__': (59, 1), "'='": (59, 1), "'/'": (59, 1), "'-'": (59, 1), "','": (59, 1), "')'": (59, 1), "'<='": (59, 1), "'>='": (59, 1), "'*'": (59, 1), "'}'": (59, 1), "'+'": (59, 1), "'<'": (59, 1)}, {}, ['expr']), ({'T.integer': 254}, {24: 157}, ['expt']), ({"'min'": 276, 'T.integer': 254, "'max'": 277, "'-'": 275, "'V'": 255, "'('": 252, "'E'": 273, "'+'": 274}, {32: 253, 34: 156, 68: 266, 69: 267, 70: 268, 71: 269, 72: 270, 73: 271, 42: 256, 56: 262, 51: 257, 52: 258, 53: 259, 54: 260, 55: 261, 24: 251, 57: 263, 58: 264, 59: 265, 74: 272}, ['expr']), ({"'min'": 276, 'T.integer': 254, "'max'": 277, "'-'": 275, "'V'": 255, "'('": 252, "'E'": 273, "'+'": 274}, {32: 253, 34: 90, 68: 266, 69: 267, 70: 268, 71: 269, 72: 270, 73: 271, 42: 256, 56: 262, 51: 257, 52: 258, 53: 259, 54: 260, 55: 261, 24: 251, 57: 263, 58: 264, 59: 265, 74: 272}, ['expr']), ({"'('": 87}, {}, ['min']), ({"'('": 86}, {}, ['max']), ({"'B'": 85, "'true'": 83, "'min'": 143, "'false'": 84, 'T.integer': 29, "'max'": 144, "'-'": 142, "'V'": 122, "'('": 217, "'E'": 140, "'!'": 82, "'+'": 141}, {24: 119, 27: 218, 32: 120, 34: 460, 37: 65, 42: 123, 43: 66, 44: 67, 45: 68, 46: 69, 47: 70, 48: 71, 49: 72, 50: 73, 51: 124, 52: 125, 53: 126, 54: 127, 55: 128, 56: 129, 57: 130, 58: 131, 59: 132, 60: 74, 61: 75, 62: 76, 63: 77, 64: 78, 65: 79, 66: 80, 67: 81, 68: 133, 69: 134, 70: 135, 71: 136, 72: 137, 73: 138, 74: 139}, ['boolexpr_p', 'expr_p']), ({"')'": 220, "'||'": 222, "'&&'": 221}, {}, ['boolexpr', 'whiles']), ({"'!='": 173, "'>'": 176, "'/'": 163, "'-'": 161, "'<='": 175, "'>='": 177, "'*'": 162, "'='": 172, "'+'": 160, "'<'": 174}, {}, ['boolexpr', 'expr']), ({"')'": (27, 1, 'boolexpr'), "'||'": (27, 1, 'boolexpr'), "'&&'": (27, 1, 'boolexpr')}, {}, ['boolexpr']), ({"')'": (37, 1), "'||'": (37, 1), "'&&'": (37, 1)}, {}, ['boolexpr']), ({"')'": (37, 1), "'||'": (37, 1), "'&&'": (37, 1)}, {}, ['boolexpr']), ({"')'": (37, 1), "'||'": (37, 1), "'&&'": (37, 1)}, {}, ['boolexpr']), ({"')'": (37, 1), "'||'": (37, 1), "'&&'": (37, 1)}, {}, ['boolexpr']), ({"')'": (37, 1), "'||'": (37, 1), "'&&'": (37, 1)}, {}, ['boolexpr']), ({"')'": (37, 1), "'||'": (37, 1), "'&&'": (37, 1)}, {}, ['boolexpr']), ({"')'": (37, 1), "'||'": (37, 1), "'&&'": (37, 1)}, {}, ['boolexpr']), ({"')'": (37, 1), "'||'": (37, 1), "'&&'": (37, 1)}, {}, ['boolexpr']), ({"')'": (43, 1), "'||'": (43, 1), "'&&'": (43, 1)}, {}, ['boolexpr']), ({"')'": (44, 1), "'||'": (44, 1), "'&&'": (44, 1)}, {}, ['boolexpr']), ({"')'": (45, 1), "'||'": (45, 1), "'&&'": (45, 1)}, {}, ['boolexpr']), ({"')'": (46, 1), "'||'": (46, 1), "'&&'": (46, 1)}, {}, ['boolexpr']), ({"')'": (47, 1), "'||'": (47, 1), "'&&'": (47, 1)}, {}, ['boolexpr']), ({"')'": (48, 1), "'||'": (48, 1), "'&&'": (48, 1)}, {}, ['boolexpr']), ({"')'": (49, 1), "'||'": (49, 1), "'&&'": (49, 1)}, {}, ['boolexpr']), ({"')'": (50, 1), "'||'": (50, 1), "'&&'": (50, 1)}, {}, ['boolexpr']), ({"'B'": 85, "'true'": 83, "'min'": 60, "'false'": 84, 'T.integer': 29, "'max'": 61, "'-'": 59, "'V'": 25, "'('": 62, "'E'": 57, "'!'": 82, "'+'": 58}, {24: 35, 27: 171, 32: 37, 34: 64, 37: 65, 42: 40, 43: 66, 44: 67, 45: 68, 46: 69, 47: 70, 48: 71, 49: 72, 50: 73, 51: 41, 52: 42, 53: 43, 54: 44, 55: 45, 56: 46, 57: 47, 58: 48, 59: 49, 60: 74, 61: 75, 62: 76, 63: 77, 64: 78, 65: 79, 66: 80, 67: 81, 68: 50, 69: 51, 70: 52, 71: 53, 72: 54, 73: 55, 74: 56}, ['boolexpr']), ({"')'": (64, 1, 'true'), "'||'": (64, 1, 'true'), "'&&'": (64, 1, 'true')}, {}, ['true']), ({"')'": (65, 1, 'false'), "'||'": (65, 1, 'false'), "'&&'": (65, 1, 'false')}, {}, ['false']), ({'T.integer': 29}, {24: 170}, ['bexpt']), ({"'min'": 60, 'T.integer': 117, "'max'": 61, "'-'": 59, "'V'": 25, "'('": 115, "'E'": 57, "'+'": 58}, {32: 37, 34: 118, 68: 50, 69: 51, 70: 52, 71: 53, 72: 54, 73: 55, 42: 40, 56: 46, 51: 41, 52: 42, 53: 43, 54: 44, 55: 45, 24: 35, 57: 47, 58: 48, 59: 49, 74: 56}, ['max']), ({"'min'": 60, 'T.integer': 117, "'max'": 61, "'-'": 59, "'V'": 25, "'('": 115, "'E'": 57, "'+'": 58}, {32: 37, 34: 145, 68: 50, 69: 51, 70: 52, 71: 53, 72: 54, 73: 55, 42: 40, 56: 46, 51: 41, 52: 42, 53: 43, 54: 44, 55: 45, 24: 35, 57: 47, 58: 48, 59: 49, 74: 56}, ['min']), ({"'!='": (51, 1), "'>'": (51, 1), 'T.newline': (51, 1), '__end_of_input__': (51, 1), "'/'": (51, 1), "'-'": (51, 1), "','": (51, 1), "')'": (51, 1), "'<='": (51, 1), "'+'": (51, 1), "'||'": (51, 1), "'>='": (51, 1), "'*'": (51, 1), "'='": (51, 1), "'&&'": (51, 1), "'<'": (51, 1)}, {}, ['expr']), ({"'!='": (52, 1), "'>'": (52, 1), 'T.newline': (52, 1), '__end_of_input__': (52, 1), "'/'": (52, 1), "'-'": (52, 1), "','": (52, 1), "')'": (52, 1), "'<='": (52, 1), "'+'": (52, 1), "'||'": (52, 1), "'>='": (52, 1), "'*'": (52, 1), "'='": (52, 1), "'&&'": (52, 1), "'<'": (52, 1)}, {}, ['expr']), ({"'!='": (69, 2), "'>'": (69, 2), 'T.newline': (69, 2), '__end_of_input__': (69, 2), "'='": (69, 2), "'/'": (69, 2), "'-'": (69, 2), "','": (69, 2), "')'": (69, 2), "'<='": (69, 2), "'>='": (69, 2), "'*'": (69, 2), "'}'": (69, 2), "'+'": (69, 2), "'<'": (69, 2)}, {}, ['expr']), ({'T.integer': 254}, {24: 455}, ['lval']), ({"'!='": (34, 1, 'expr'), "'>'": (34, 1, 'expr'), 'T.newline': (34, 1, 'expr'), '__end_of_input__': (34, 1, 'expr'), "'/'": (34, 1, 'expr'), "'-'": (34, 1, 'expr'), "','": (34, 1, 'expr'), "')'": (34, 1, 'expr'), "'<='": (34, 1, 'expr'), "'+'": (34, 1, 'expr'), "'||'": (34, 1, 'expr'), "'>='": (34, 1, 'expr'), "'*'": (34, 1, 'expr'), "'='": (34, 1, 'expr'), "'&&'": (34, 1, 'expr'), "'<'": (34, 1, 'expr')}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'='": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'='": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'='": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'='": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'='": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'='": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'='": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'='": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'='": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (53, 1), "'>'": (53, 1), 'T.newline': (53, 1), '__end_of_input__': (53, 1), "'/'": (53, 1), "'-'": (53, 1), "','": (53, 1), "')'": (53, 1), "'<='": (53, 1), "'+'": (53, 1), "'||'": (53, 1), "'>='": (53, 1), "'*'": (53, 1), "'='": (53, 1), "'&&'": (53, 1), "'<'": (53, 1)}, {}, ['expr']), ({"'!='": (54, 1), "'>'": (54, 1), 'T.newline': (54, 1), '__end_of_input__': (54, 1), "'/'": (54, 1), "'-'": (54, 1), "','": (54, 1), "')'": (54, 1), "'<='": (54, 1), "'+'": (54, 1), "'||'": (54, 1), "'>='": (54, 1), "'*'": (54, 1), "'='": (54, 1), "'&&'": (54, 1), "'<'": (54, 1)}, {}, ['expr']), ({"'!='": (55, 1), "'>'": (55, 1), 'T.newline': (55, 1), '__end_of_input__': (55, 1), "'/'": (55, 1), "'-'": (55, 1), "','": (55, 1), "')'": (55, 1), "'<='": (55, 1), "'+'": (55, 1), "'||'": (55, 1), "'>='": (55, 1), "'*'": (55, 1), "'='": (55, 1), "'&&'": (55, 1), "'<'": (55, 1)}, {}, ['expr']), ({"'!='": (56, 1), "'>'": (56, 1), 'T.newline': (56, 1), '__end_of_input__': (56, 1), "'/'": (56, 1), "'-'": (56, 1), "','": (56, 1), "')'": (56, 1), "'<='": (56, 1), "'+'": (56, 1), "'||'": (56, 1), "'>='": (56, 1), "'*'": (56, 1), "'='": (56, 1), "'&&'": (56, 1), "'<'": (56, 1)}, {}, ['expr']), ({"'!='": (57, 1), "'>'": (57, 1), 'T.newline': (57, 1), '__end_of_input__': (57, 1), "'/'": (57, 1), "'-'": (57, 1), "','": (57, 1), "')'": (57, 1), "'<='": (57, 1), "'+'": (57, 1), "'||'": (57, 1), "'>='": (57, 1), "'*'": (57, 1), "'='": (57, 1), "'&&'": (57, 1), "'<'": (57, 1)}, {}, ['expr']), ({"'!='": (58, 1), "'>'": (58, 1), 'T.newline': (58, 1), '__end_of_input__': (58, 1), "'/'": (58, 1), "'-'": (58, 1), "','": (58, 1), "')'": (58, 1), "'<='": (58, 1), "'+'": (58, 1), "'||'": (58, 1), "'>='": (58, 1), "'*'": (58, 1), "'='": (58, 1), "'&&'": (58, 1), "'<'": (58, 1)}, {}, ['expr']), ({"'!='": (59, 1), "'>'": (59, 1), 'T.newline': (59, 1), '__end_of_input__': (59, 1), "'/'": (59, 1), "'-'": (59, 1), "','": (59, 1), "')'": (59, 1), "'<='": (59, 1), "'+'": (59, 1), "'||'": (59, 1), "'>='": (59, 1), "'*'": (59, 1), "'='": (59, 1), "'&&'": (59, 1), "'<'": (59, 1)}, {}, ['expr']), ({'T.integer': 254}, {24: 150}, ['expt']), ({"'min'": 276, 'T.integer': 254, "'max'": 277, "'-'": 275, "'V'": 255, "'('": 252, "'E'": 273, "'+'": 274}, {32: 253, 34: 149, 68: 266, 69: 267, 70: 268, 71: 269, 72: 270, 73: 271, 42: 256, 56: 262, 51: 257, 52: 258, 53: 259, 54: 260, 55: 261, 24: 251, 57: 263, 58: 264, 59: 265, 74: 272}, ['expr']), ({"'min'": 276, 'T.integer': 254, "'max'": 277, "'-'": 275, "'V'": 255, "'('": 252, "'E'": 273, "'+'": 274}, {32: 253, 34: 148, 68: 266, 69: 267, 70: 268, 71: 269, 72: 270, 73: 271, 42: 256, 56: 262, 51: 257, 52: 258, 53: 259, 54: 260, 55: 261, 24: 251, 57: 263, 58: 264, 59: 265, 74: 272}, ['expr']), ({"'('": 147}, {}, ['min']), ({"'('": 146}, {}, ['max']), ({"')'": 155, "'*'": 153, "'/'": 154, "'-'": 152, "'+'": 151}, {}, ['expr', 'expr_p']), ({"'min'": 143, 'T.integer': 29, "'max'": 144, "'-'": 142, "'V'": 122, "'('": 115, "'E'": 140, "'+'": 141}, {32: 120, 34: 121, 68: 133, 69: 134, 70: 135, 71: 136, 72: 137, 73: 138, 42: 123, 56: 129, 51: 124, 52: 125, 53: 126, 54: 127, 55: 128, 24: 119, 57: 130, 58: 131, 59: 132, 74: 139}, ['expr_p']), ({"'*'": 401, "'/'": 402, "'-'": 400, "'+'": 399, "','": 536}, {}, ['expr', 'max']), ({"'/'": (24, 1, 'num'), "'-'": (24, 1, 'num'), "','": (24, 1, 'num'), "')'": (24, 1, 'num'), "'*'": (24, 1, 'num'), "'+'": (24, 1, 'num')}, {}, ['num']), ({"'*'": 162, "'/'": 163, "'-'": 161, "'+'": 160, "','": 164}, {}, ['expr', 'max']), ({"'!='": (51, 1), "'>'": (51, 1), 'T.newline': (51, 1), "'='": (51, 1), "'/'": (51, 1), "'-'": (51, 1), "','": (51, 1), "')'": (51, 1), "'<='": (51, 1), "'>='": (51, 1), "'*'": (51, 1), "'}'": (51, 1), "'+'": (51, 1), "'<'": (51, 1)}, {}, ['expr']), ({"'!='": (52, 1), "'>'": (52, 1), 'T.newline': (52, 1), "'='": (52, 1), "'/'": (52, 1), "'-'": (52, 1), "','": (52, 1), "')'": (52, 1), "'<='": (52, 1), "'>='": (52, 1), "'*'": (52, 1), "'}'": (52, 1), "'+'": (52, 1), "'<'": (52, 1)}, {}, ['expr']), ({"')'": 223, "'*'": 162, "'/'": 163, "'-'": 161, "'+'": 160}, {}, ['expr', 'expr_p']), ({'T.integer': 179}, {24: 212}, ['lval']), ({"'!='": (34, 1, 'expr'), "'>'": (34, 1, 'expr'), 'T.newline': (34, 1, 'expr'), "'='": (34, 1, 'expr'), "'/'": (34, 1, 'expr'), "'-'": (34, 1, 'expr'), "','": (34, 1, 'expr'), "')'": (34, 1, 'expr'), "'<='": (34, 1, 'expr'), "'>='": (34, 1, 'expr'), "'*'": (34, 1, 'expr'), "'}'": (34, 1, 'expr'), "'+'": (34, 1, 'expr'), "'<'": (34, 1, 'expr')}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (53, 1), "'>'": (53, 1), 'T.newline': (53, 1), "'='": (53, 1), "'/'": (53, 1), "'-'": (53, 1), "','": (53, 1), "')'": (53, 1), "'<='": (53, 1), "'>='": (53, 1), "'*'": (53, 1), "'}'": (53, 1), "'+'": (53, 1), "'<'": (53, 1)}, {}, ['expr']), ({"'!='": (54, 1), "'>'": (54, 1), 'T.newline': (54, 1), "'='": (54, 1), "'/'": (54, 1), "'-'": (54, 1), "','": (54, 1), "')'": (54, 1), "'<='": (54, 1), "'>='": (54, 1), "'*'": (54, 1), "'}'": (54, 1), "'+'": (54, 1), "'<'": (54, 1)}, {}, ['expr']), ({"'!='": (55, 1), "'>'": (55, 1), 'T.newline': (55, 1), "'='": (55, 1), "'/'": (55, 1), "'-'": (55, 1), "','": (55, 1), "')'": (55, 1), "'<='": (55, 1), "'>='": (55, 1), "'*'": (55, 1), "'}'": (55, 1), "'+'": (55, 1), "'<'": (55, 1)}, {}, ['expr']), ({"'!='": (56, 1), "'>'": (56, 1), 'T.newline': (56, 1), "'='": (56, 1), "'/'": (56, 1), "'-'": (56, 1), "','": (56, 1), "')'": (56, 1), "'<='": (56, 1), "'>='": (56, 1), "'*'": (56, 1), "'}'": (56, 1), "'+'": (56, 1), "'<'": (56, 1)}, {}, ['expr']), ({"'!='": (57, 1), "'>'": (57, 1), 'T.newline': (57, 1), "'='": (57, 1), "'/'": (57, 1), "'-'": (57, 1), "','": (57, 1), "')'": (57, 1), "'<='": (57, 1), "'>='": (57, 1), "'*'": (57, 1), "'}'": (57, 1), "'+'": (57, 1), "'<'": (57, 1)}, {}, ['expr']), ({"'!='": (58, 1), "'>'": (58, 1), 'T.newline': (58, 1), "'='": (58, 1), "'/'": (58, 1), "'-'": (58, 1), "','": (58, 1), "')'": (58, 1), "'<='": (58, 1), "'>='": (58, 1), "'*'": (58, 1), "'}'": (58, 1), "'+'": (58, 1), "'<'": (58, 1)}, {}, ['expr']), ({"'!='": (59, 1), "'>'": (59, 1), 'T.newline': (59, 1), "'='": (59, 1), "'/'": (59, 1), "'-'": (59, 1), "','": (59, 1), "')'": (59, 1), "'<='": (59, 1), "'>='": (59, 1), "'*'": (59, 1), "'}'": (59, 1), "'+'": (59, 1), "'<'": (59, 1)}, {}, ['expr']), ({'T.integer': 179}, {24: 211}, ['expt']), ({"'min'": 486, 'T.integer': 179, "'max'": 487, "'-'": 485, "'V'": 465, "'('": 178, "'E'": 483, "'+'": 484}, {32: 464, 34: 210, 68: 476, 69: 477, 70: 478, 71: 479, 72: 480, 73: 481, 42: 466, 56: 472, 51: 467, 52: 468, 53: 469, 54: 470, 55: 471, 24: 463, 57: 473, 58: 474, 59: 475, 74: 482}, ['expr']), ({"'min'": 486, 'T.integer': 179, "'max'": 487, "'-'": 485, "'V'": 465, "'('": 178, "'E'": 483, "'+'": 484}, {32: 464, 34: 182, 68: 476, 69: 477, 70: 478, 71: 479, 72: 480, 73: 481, 42: 466, 56: 472, 51: 467, 52: 468, 53: 469, 54: 470, 55: 471, 24: 463, 57: 473, 58: 474, 59: 475, 74: 482}, ['expr']), ({"'('": 181}, {}, ['min']), ({"'('": 180}, {}, ['max']), ({"'*'": 162, "'/'": 163, "'-'": 161, "'+'": 160, "','": 165}, {}, ['expr', 'min']), ({"'min'": 60, 'T.integer': 117, "'max'": 61, "'-'": 59, "'V'": 25, "'('": 115, "'E'": 57, "'+'": 58}, {32: 37, 34: 228, 68: 50, 69: 51, 70: 52, 71: 53, 72: 54, 73: 55, 42: 40, 56: 46, 51: 41, 52: 42, 53: 43, 54: 44, 55: 45, 24: 35, 57: 47, 58: 48, 59: 49, 74: 56}, ['max']), ({"'min'": 60, 'T.integer': 117, "'max'": 61, "'-'": 59, "'V'": 25, "'('": 115, "'E'": 57, "'+'": 58}, {32: 37, 34: 229, 68: 50, 69: 51, 70: 52, 71: 53, 72: 54, 73: 55, 42: 40, 56: 46, 51: 41, 52: 42, 53: 43, 54: 44, 55: 45, 24: 35, 57: 47, 58: 48, 59: 49, 74: 56}, ['min']), ({"'!='": (69, 2), "'>'": (69, 2), 'T.newline': (69, 2), '__end_of_input__': (69, 2), "'/'": (69, 2), "'-'": (69, 2), "','": (69, 2), "')'": (69, 2), "'<='": (69, 2), "'+'": (69, 2), "'||'": (69, 2), "'>='": (69, 2), "'*'": (69, 2), "'='": (69, 2), "'&&'": (69, 2), "'<'": (69, 2)}, {}, ['expr']), ({"'!='": (69, 2), "'>'": (69, 2), 'T.newline': (69, 2), '__end_of_input__': (69, 2), "'/'": (69, 2), "'-'": (69, 2), "','": (69, 2), "')'": (69, 2), "'<='": (69, 2), "'+'": (69, 2), "'||'": (69, 2), "'>='": (69, 2), "'*'": (69, 2), "'='": (69, 2), "'&&'": (69, 2), "'<'": (69, 2)}, {}, ['expr']), ({"'!='": (68, 2, 'expt'), "'>'": (68, 2, 'expt'), 'T.newline': (68, 2, 'expt'), '__end_of_input__': (68, 2, 'expt'), "'/'": (68, 2, 'expt'), "'-'": (68, 2, 'expt'), "','": (68, 2, 'expt'), "')'": (68, 2, 'expt'), "'<='": (68, 2, 'expt'), "'+'": (68, 2, 'expt'), "'||'": (68, 2, 'expt'), "'>='": (68, 2, 'expt'), "'*'": (68, 2, 'expt'), "'='": (68, 2, 'expt'), "'&&'": (68, 2, 'expt'), "'<'": (68, 2, 'expt')}, {}, ['expt']), ({"'min'": 276, 'T.integer': 254, "'max'": 277, "'-'": 275, "'V'": 255, "'('": 252, "'E'": 273, "'+'": 274}, {32: 253, 34: 230, 68: 266, 69: 267, 70: 268, 71: 269, 72: 270, 73: 271, 42: 256, 56: 262, 51: 257, 52: 258, 53: 259, 54: 260, 55: 261, 24: 251, 57: 263, 58: 264, 59: 265, 74: 272}, ['expr']), ({"'min'": 276, 'T.integer': 254, "'max'": 277, "'-'": 275, "'V'": 255, "'('": 252, "'E'": 273, "'+'": 274}, {32: 253, 34: 231, 68: 266, 69: 267, 70: 268, 71: 269, 72: 270, 73: 271, 42: 256, 56: 262, 51: 257, 52: 258, 53: 259, 54: 260, 55: 261, 24: 251, 57: 263, 58: 264, 59: 265, 74: 272}, ['expr']), ({"'min'": 276, 'T.integer': 254, "'max'": 277, "'-'": 275, "'V'": 255, "'('": 252, "'E'": 273, "'+'": 274}, {32: 253, 34: 232, 68: 266, 69: 267, 70: 268, 71: 269, 72: 270, 73: 271, 42: 256, 56: 262, 51: 257, 52: 258, 53: 259, 54: 260, 55: 261, 24: 251, 57: 263, 58: 264, 59: 265, 74: 272}, ['expr']), ({"'min'": 276, 'T.integer': 254, "'max'": 277, "'-'": 275, "'V'": 255, "'('": 252, "'E'": 273, "'+'": 274}, {32: 253, 34: 233, 68: 266, 69: 267, 70: 268, 71: 269, 72: 270, 73: 271, 42: 256, 56: 262, 51: 257, 52: 258, 53: 259, 54: 260, 55: 261, 24: 251, 57: 263, 58: 264, 59: 265, 74: 272}, ['expr']), ({"'!='": (72, 3, 'expr_p'), "'>'": (72, 3, 'expr_p'), 'T.newline': (72, 3, 'expr_p'), '__end_of_input__': (72, 3, 'expr_p'), "'='": (72, 3, 'expr_p'), "'/'": (72, 3, 'expr_p'), "'-'": (72, 3, 'expr_p'), "','": (72, 3, 'expr_p'), "')'": (72, 3, 'expr_p'), "'<='": (72, 3, 'expr_p'), "'+'": (72, 3, 'expr_p'), "'||'": (72, 3, 'expr_p'), "'>='": (72, 3, 'expr_p'), "'*'": (72, 3, 'expr_p'), "'}'": (72, 3, 'expr_p'), "'&&'": (72, 3, 'expr_p'), "'<'": (72, 3, 'expr_p')}, {}, ['expr_p']), ({"'!='": (69, 2), "'>'": (69, 2), 'T.newline': (69, 2), '__end_of_input__': (69, 2), "'='": (69, 2), "'/'": (69, 2), "'-'": (69, 2), "','": (69, 2), "')'": (69, 2), "'<='": (69, 2), "'>='": (69, 2), "'*'": (69, 2), "'}'": (69, 2), "'+'": (69, 2), "'<'": (69, 2)}, {}, ['expr']), ({"'!='": (68, 2, 'expt'), "'>'": (68, 2, 'expt'), 'T.newline': (68, 2, 'expt'), '__end_of_input__': (68, 2, 'expt'), "'='": (68, 2, 'expt'), "'/'": (68, 2, 'expt'), "'-'": (68, 2, 'expt'), "','": (68, 2, 'expt'), "')'": (68, 2, 'expt'), "'<='": (68, 2, 'expt'), "'>='": (68, 2, 'expt'), "'*'": (68, 2, 'expt'), "'}'": (68, 2, 'expt'), "'+'": (68, 2, 'expt'), "'<'": (68, 2, 'expt')}, {}, ['expt']), ({"'!='": (32, 2, 'lval'), "'>'": (32, 2, 'lval'), 'T.newline': (32, 2, 'lval'), '__end_of_input__': (32, 2, 'lval'), "'='": (32, 2, 'lval'), "'/'": (32, 2, 'lval'), "'-'": (32, 2, 'lval'), "','": (32, 2, 'lval'), "')'": (32, 2, 'lval'), "'<='": (32, 2, 'lval'), "':='": (32, 2, 'lval'), "'>='": (32, 2, 'lval'), "'*'": (32, 2, 'lval'), "'}'": (32, 2, 'lval'), "'+'": (32, 2, 'lval'), "'<'": (32, 2, 'lval')}, {}, ['lval']), ({"')'": 234, "'||'": 222, "'&&'": 221}, {}, ['boolexpr', 'ifs']), ({"'min'": 486, 'T.integer': 179, "'max'": 487, "'-'": 485, "'V'": 465, "'('": 178, "'E'": 483, "'+'": 484}, {32: 464, 34: 169, 68: 476, 69: 477, 70: 478, 71: 479, 72: 480, 73: 481, 42: 466, 56: 472, 51: 467, 52: 468, 53: 469, 54: 470, 55: 471, 24: 463, 57: 473, 58: 474, 59: 475, 74: 482}, ['expr']), ({"'min'": 486, 'T.integer': 179, "'max'": 487, "'-'": 485, "'V'": 465, "'('": 178, "'E'": 483, "'+'": 484}, {32: 464, 34: 168, 68: 476, 69: 477, 70: 478, 71: 479, 72: 480, 73: 481, 42: 466, 56: 472, 51: 467, 52: 468, 53: 469, 54: 470, 55: 471, 24: 463, 57: 473, 58: 474, 59: 475, 74: 482}, ['expr']), ({"'min'": 486, 'T.integer': 179, "'max'": 487, "'-'": 485, "'V'": 465, "'('": 178, "'E'": 483, "'+'": 484}, {32: 464, 34: 167, 68: 476, 69: 477, 70: 478, 71: 479, 72: 480, 73: 481, 42: 466, 56: 472, 51: 467, 52: 468, 53: 469, 54: 470, 55: 471, 24: 463, 57: 473, 58: 474, 59: 475, 74: 482}, ['expr']), ({"'min'": 486, 'T.integer': 179, "'max'": 487, "'-'": 485, "'V'": 465, "'('": 178, "'E'": 483, "'+'": 484}, {32: 464, 34: 166, 68: 476, 69: 477, 70: 478, 71: 479, 72: 480, 73: 481, 42: 466, 56: 472, 51: 467, 52: 468, 53: 469, 54: 470, 55: 471, 24: 463, 57: 473, 58: 474, 59: 475, 74: 482}, ['expr']), ({"'min'": 208, 'T.integer': 117, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 184, "'E'": 205, "'+'": 206}, {32: 185, 34: 237, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['max']), ({"'min'": 208, 'T.integer': 117, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 184, "'E'": 205, "'+'": 206}, {32: 185, 34: 247, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['min']), ({"'!='": (70, 3), "'>'": (70, 3), 'T.newline': (70, 3), "'='": (70, 3), "'/'": (70, 3), "'-'": (70, 3), "','": (70, 3), "')'": (70, 3), "'<='": (70, 3), "'>='": (70, 3), "'*'": (70, 3), "'}'": (70, 3), "'+'": (70, 3), "'<'": (70, 3)}, {}, ['expr']), ({"'!='": (70, 3), "'>'": (70, 3), 'T.newline': (70, 3), "'='": (70, 3), "'/'": (70, 3), "'-'": (70, 3), "','": (70, 3), "')'": (70, 3), "'<='": (70, 3), "'>='": (70, 3), "'*'": (70, 3), "'}'": (70, 3), "'+'": (70, 3), "'<'": (70, 3)}, {}, ['expr']), ({"'!='": (71, 3), "'>'": (71, 3), 'T.newline': (71, 3), "'='": (71, 3), "'/'": (71, 3), "'-'": (71, 3), "','": (71, 3), "')'": (71, 3), "'<='": (71, 3), "'>='": (71, 3), "'*'": (71, 3), "'}'": (71, 3), "'+'": (71, 3), "'<'": (71, 3)}, {}, ['expr']), ({"'!='": (71, 3), "'>'": (71, 3), 'T.newline': (71, 3), "'='": (71, 3), "'/'": (71, 3), "'-'": (71, 3), "','": (71, 3), "')'": (71, 3), "'<='": (71, 3), "'>='": (71, 3), "'*'": (71, 3), "'}'": (71, 3), "'+'": (71, 3), "'<'": (71, 3)}, {}, ['expr']), ({"')'": (66, 2, 'bexpt'), "'||'": (66, 2, 'bexpt'), "'&&'": (66, 2, 'bexpt')}, {}, ['bexpt']), ({"')'": (60, 2), "'||'": (60, 2), "'&&'": (60, 2)}, {}, ['boolexpr']), ({"'min'": 112, 'T.integer': 39, "'max'": 113, "'-'": 111, "'V'": 91, "'('": 36, "'E'": 109, "'+'": 110}, {32: 89, 34: 248, 68: 102, 69: 103, 70: 104, 71: 105, 72: 106, 73: 107, 42: 92, 56: 98, 51: 93, 52: 94, 53: 95, 54: 96, 55: 97, 24: 88, 57: 99, 58: 100, 59: 101, 74: 108}, ['boolexpr']), ({"'min'": 112, 'T.integer': 39, "'max'": 113, "'-'": 111, "'V'": 91, "'('": 36, "'E'": 109, "'+'": 110}, {32: 89, 34: 278, 68: 102, 69: 103, 70: 104, 71: 105, 72: 106, 73: 107, 42: 92, 56: 98, 51: 93, 52: 94, 53: 95, 54: 96, 55: 97, 24: 88, 57: 99, 58: 100, 59: 101, 74: 108}, ['boolexpr']), ({"'min'": 112, 'T.integer': 39, "'max'": 113, "'-'": 111, "'V'": 91, "'('": 36, "'E'": 109, "'+'": 110}, {32: 89, 34: 279, 68: 102, 69: 103, 70: 104, 71: 105, 72: 106, 73: 107, 42: 92, 56: 98, 51: 93, 52: 94, 53: 95, 54: 96, 55: 97, 24: 88, 57: 99, 58: 100, 59: 101, 74: 108}, ['boolexpr']), ({"'min'": 112, 'T.integer': 39, "'max'": 113, "'-'": 111, "'V'": 91, "'('": 36, "'E'": 109, "'+'": 110}, {32: 89, 34: 280, 68: 102, 69: 103, 70: 104, 71: 105, 72: 106, 73: 107, 42: 92, 56: 98, 51: 93, 52: 94, 53: 95, 54: 96, 55: 97, 24: 88, 57: 99, 58: 100, 59: 101, 74: 108}, ['boolexpr']), ({"'min'": 112, 'T.integer': 39, "'max'": 113, "'-'": 111, "'V'": 91, "'('": 36, "'E'": 109, "'+'": 110}, {32: 89, 34: 281, 68: 102, 69: 103, 70: 104, 71: 105, 72: 106, 73: 107, 42: 92, 56: 98, 51: 93, 52: 94, 53: 95, 54: 96, 55: 97, 24: 88, 57: 99, 58: 100, 59: 101, 74: 108}, ['boolexpr']), ({"'min'": 112, 'T.integer': 39, "'max'": 113, "'-'": 111, "'V'": 91, "'('": 36, "'E'": 109, "'+'": 110}, {32: 89, 34: 282, 68: 102, 69: 103, 70: 104, 71: 105, 72: 106, 73: 107, 42: 92, 56: 98, 51: 93, 52: 94, 53: 95, 54: 96, 55: 97, 24: 88, 57: 99, 58: 100, 59: 101, 74: 108}, ['boolexpr']), ({"'min'": 208, 'T.integer': 117, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 184, "'E'": 205, "'+'": 206}, {32: 185, 34: 186, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['expr_p']), ({"'!='": (24, 1, 'num'), "'>'": (24, 1, 'num'), 'T.newline': (24, 1, 'num'), "'='": (24, 1, 'num'), "'/'": (24, 1, 'num'), "'-'": (24, 1, 'num'), "','": (24, 1, 'num'), "')'": (24, 1, 'num'), "'<='": (24, 1, 'num'), "'>='": (24, 1, 'num'), "'*'": (24, 1, 'num'), "'}'": (24, 1, 'num'), "'+'": (24, 1, 'num'), "'<'": (24, 1, 'num')}, {}, ['num']), ({"'min'": 208, 'T.integer': 284, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 115, "'E'": 205, "'+'": 206}, {32: 185, 34: 283, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['max']), ({"'min'": 208, 'T.integer': 284, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 115, "'E'": 205, "'+'": 206}, {32: 185, 34: 329, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['min']), ({"'!='": (69, 2), "'>'": (69, 2), 'T.newline': (69, 2), "'='": (69, 2), "'/'": (69, 2), "'-'": (69, 2), "','": (69, 2), "')'": (69, 2), "'<='": (69, 2), "'>='": (69, 2), "'*'": (69, 2), "'}'": (69, 2), "'+'": (69, 2), "'<'": (69, 2)}, {}, ['expr']), ({"'/'": (51, 1), "'-'": (51, 1), "','": (51, 1), "')'": (51, 1), "'*'": (51, 1), "'+'": (51, 1)}, {}, ['expr']), ({"'min'": 112, 'T.integer': 39, "'max'": 113, "'-'": 111, "'V'": 91, "'('": 36, "'E'": 109, "'+'": 110}, {32: 89, 34: 246, 68: 102, 69: 103, 70: 104, 71: 105, 72: 106, 73: 107, 42: 92, 56: 98, 51: 93, 52: 94, 53: 95, 54: 96, 55: 97, 24: 88, 57: 99, 58: 100, 59: 101, 74: 108}, ['expr_p']), ({"'/'": (52, 1), "'-'": (52, 1), "','": (52, 1), "')'": (52, 1), "'*'": (52, 1), "'+'": (52, 1)}, {}, ['expr']), ({"')'": 330, "'*'": 401, "'/'": 402, "'-'": 400, "'+'": 399}, {}, ['expr', 'expr_p']), ({'T.integer': 290}, {24: 245}, ['lval']), ({"'/'": (34, 1, 'expr'), "'-'": (34, 1, 'expr'), "','": (34, 1, 'expr'), "')'": (34, 1, 'expr'), "'*'": (34, 1, 'expr'), "'+'": (34, 1, 'expr')}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (53, 1), "'-'": (53, 1), "','": (53, 1), "')'": (53, 1), "'*'": (53, 1), "'+'": (53, 1)}, {}, ['expr']), ({"'/'": (54, 1), "'-'": (54, 1), "','": (54, 1), "')'": (54, 1), "'*'": (54, 1), "'+'": (54, 1)}, {}, ['expr']), ({"'/'": (55, 1), "'-'": (55, 1), "','": (55, 1), "')'": (55, 1), "'*'": (55, 1), "'+'": (55, 1)}, {}, ['expr']), ({"'/'": (56, 1), "'-'": (56, 1), "','": (56, 1), "')'": (56, 1), "'*'": (56, 1), "'+'": (56, 1)}, {}, ['expr']), ({"'/'": (57, 1), "'-'": (57, 1), "','": (57, 1), "')'": (57, 1), "'*'": (57, 1), "'+'": (57, 1)}, {}, ['expr']), ({"'/'": (58, 1), "'-'": (58, 1), "','": (58, 1), "')'": (58, 1), "'*'": (58, 1), "'+'": (58, 1)}, {}, ['expr']), ({"'/'": (59, 1), "'-'": (59, 1), "','": (59, 1), "')'": (59, 1), "'*'": (59, 1), "'+'": (59, 1)}, {}, ['expr']), ({'T.integer': 290}, {24: 244}, ['expt']), ({"'min'": 312, 'T.integer': 290, "'max'": 313, "'-'": 311, "'V'": 291, "'('": 288, "'E'": 309, "'+'": 310}, {32: 289, 34: 243, 68: 302, 69: 303, 70: 304, 71: 305, 72: 306, 73: 307, 42: 292, 56: 298, 51: 293, 52: 294, 53: 295, 54: 296, 55: 297, 24: 287, 57: 299, 58: 300, 59: 301, 74: 308}, ['expr']), ({"'min'": 312, 'T.integer': 290, "'max'": 313, "'-'": 311, "'V'": 291, "'('": 288, "'E'": 309, "'+'": 310}, {32: 289, 34: 240, 68: 302, 69: 303, 70: 304, 71: 305, 72: 306, 73: 307, 42: 292, 56: 298, 51: 293, 52: 294, 53: 295, 54: 296, 55: 297, 24: 287, 57: 299, 58: 300, 59: 301, 74: 308}, ['expr']), ({"'('": 239}, {}, ['min']), ({"'('": 238}, {}, ['max']), ({"'!='": (69, 2), "'>'": (69, 2), 'T.newline': (69, 2), "'='": (69, 2), "'/'": (69, 2), "'-'": (69, 2), "','": (69, 2), "')'": (69, 2), "'<='": (69, 2), "'>='": (69, 2), "'*'": (69, 2), "'}'": (69, 2), "'+'": (69, 2), "'<'": (69, 2)}, {}, ['expr']), ({"'!='": (68, 2, 'expt'), "'>'": (68, 2, 'expt'), 'T.newline': (68, 2, 'expt'), "'='": (68, 2, 'expt'), "'/'": (68, 2, 'expt'), "'-'": (68, 2, 'expt'), "','": (68, 2, 'expt'), "')'": (68, 2, 'expt'), "'<='": (68, 2, 'expt'), "'>='": (68, 2, 'expt'), "'*'": (68, 2, 'expt'), "'}'": (68, 2, 'expt'), "'+'": (68, 2, 'expt'), "'<'": (68, 2, 'expt')}, {}, ['expt']), ({"'!='": (32, 2, 'lval'), "'>'": (32, 2, 'lval'), 'T.newline': (32, 2, 'lval'), "'='": (32, 2, 'lval'), "'/'": (32, 2, 'lval'), "'-'": (32, 2, 'lval'), "','": (32, 2, 'lval'), "')'": (32, 2, 'lval'), "'<='": (32, 2, 'lval'), "'>='": (32, 2, 'lval'), "'*'": (32, 2, 'lval'), "'}'": (32, 2, 'lval'), "'+'": (32, 2, 'lval'), "'<'": (32, 2, 'lval')}, {}, ['lval']), ({"'min'": 486, 'T.integer': 179, "'max'": 487, "'-'": 485, "'V'": 465, "'('": 178, "'E'": 483, "'+'": 484}, {32: 464, 34: 227, 68: 476, 69: 477, 70: 478, 71: 479, 72: 480, 73: 481, 42: 466, 56: 472, 51: 467, 52: 468, 53: 469, 54: 470, 55: 471, 24: 463, 57: 473, 58: 474, 59: 475, 74: 482}, ['expr']), ({"'min'": 486, 'T.integer': 179, "'max'": 487, "'-'": 485, "'V'": 465, "'('": 178, "'E'": 483, "'+'": 484}, {32: 464, 34: 226, 68: 476, 69: 477, 70: 478, 71: 479, 72: 480, 73: 481, 42: 466, 56: 472, 51: 467, 52: 468, 53: 469, 54: 470, 55: 471, 24: 463, 57: 473, 58: 474, 59: 475, 74: 482}, ['expr']), ({"'min'": 486, 'T.integer': 179, "'max'": 487, "'-'": 485, "'V'": 465, "'('": 178, "'E'": 483, "'+'": 484}, {32: 464, 34: 225, 68: 476, 69: 477, 70: 478, 71: 479, 72: 480, 73: 481, 42: 466, 56: 472, 51: 467, 52: 468, 53: 469, 54: 470, 55: 471, 24: 463, 57: 473, 58: 474, 59: 475, 74: 482}, ['expr']), ({"'min'": 486, 'T.integer': 179, "'max'": 487, "'-'": 485, "'V'": 465, "'('": 178, "'E'": 483, "'+'": 484}, {32: 464, 34: 224, 68: 476, 69: 477, 70: 478, 71: 479, 72: 480, 73: 481, 42: 466, 56: 472, 51: 467, 52: 468, 53: 469, 54: 470, 55: 471, 24: 463, 57: 473, 58: 474, 59: 475, 74: 482}, ['expr']), ({"'B'": 85, "'true'": 83, "'min'": 143, "'false'": 84, 'T.integer': 29, "'max'": 144, "'-'": 142, "'V'": 122, "'('": 217, "'E'": 140, "'!'": 82, "'+'": 141}, {24: 119, 27: 218, 32: 120, 34: 219, 37: 65, 42: 123, 43: 66, 44: 67, 45: 68, 46: 69, 47: 70, 48: 71, 49: 72, 50: 73, 51: 124, 52: 125, 53: 126, 54: 127, 55: 128, 56: 129, 57: 130, 58: 131, 59: 132, 60: 74, 61: 75, 62: 76, 63: 77, 64: 78, 65: 79, 66: 80, 67: 81, 68: 133, 69: 134, 70: 135, 71: 136, 72: 137, 73: 138, 74: 139}, ['boolexpr_p', 'expr_p']), ({"')'": 362, "'||'": 222, "'&&'": 221}, {}, ['boolexpr', 'boolexpr_p']), ({"'!='": 173, "'>'": 176, "'/'": 163, "'-'": 161, "')'": 223, "'<='": 175, "'>='": 177, "'*'": 162, "'='": 172, "'+'": 160, "'<'": 174}, {}, ['boolexpr', 'expr', 'expr_p']), ({"'{'": 364}, {29: 363}, ['whiles']), ({"'B'": 85, "'true'": 83, "'min'": 60, "'false'": 84, 'T.integer': 29, "'max'": 61, "'-'": 59, "'V'": 25, "'('": 62, "'E'": 57, "'!'": 82, "'+'": 58}, {24: 35, 27: 236, 32: 37, 34: 64, 37: 65, 42: 40, 43: 66, 44: 67, 45: 68, 46: 69, 47: 70, 48: 71, 49: 72, 50: 73, 51: 41, 52: 42, 53: 43, 54: 44, 55: 45, 56: 46, 57: 47, 58: 48, 59: 49, 60: 74, 61: 75, 62: 76, 63: 77, 64: 78, 65: 79, 66: 80, 67: 81, 68: 50, 69: 51, 70: 52, 71: 53, 72: 54, 73: 55, 74: 56}, ['boolexpr']), ({"'B'": 85, "'true'": 83, "'min'": 60, "'false'": 84, 'T.integer': 29, "'max'": 61, "'-'": 59, "'V'": 25, "'('": 62, "'E'": 57, "'!'": 82, "'+'": 58}, {24: 35, 27: 235, 32: 37, 34: 64, 37: 65, 42: 40, 43: 66, 44: 67, 45: 68, 46: 69, 47: 70, 48: 71, 49: 72, 50: 73, 51: 41, 52: 42, 53: 43, 54: 44, 55: 45, 56: 46, 57: 47, 58: 48, 59: 49, 60: 74, 61: 75, 62: 76, 63: 77, 64: 78, 65: 79, 66: 80, 67: 81, 68: 50, 69: 51, 70: 52, 71: 53, 72: 54, 73: 55, 74: 56}, ['boolexpr']), ({"'!='": (72, 3, 'expr_p'), "'>'": (72, 3, 'expr_p'), 'T.newline': (72, 3, 'expr_p'), "'='": (72, 3, 'expr_p'), "'/'": (72, 3, 'expr_p'), "'-'": (72, 3, 'expr_p'), "','": (72, 3, 'expr_p'), "')'": (72, 3, 'expr_p'), "'<='": (72, 3, 'expr_p'), "'>='": (72, 3, 'expr_p'), "'*'": (72, 3, 'expr_p'), "'}'": (72, 3, 'expr_p'), "'+'": (72, 3, 'expr_p'), "'<'": (72, 3, 'expr_p')}, {}, ['expr_p']), ({"'!='": (70, 3), "'>'": (70, 3), 'T.newline': (70, 3), "'='": (70, 3), "'/'": (70, 3), "'-'": (70, 3), "','": (70, 3), "')'": (70, 3), "'<='": (70, 3), "'>='": (70, 3), "'*'": (70, 3), "'}'": (70, 3), "'+'": (70, 3), "'<'": (70, 3)}, {}, ['expr']), ({"'!='": (70, 3), "'>'": (70, 3), 'T.newline': (70, 3), "'='": (70, 3), "'/'": (70, 3), "'-'": (70, 3), "','": (70, 3), "')'": (70, 3), "'<='": (70, 3), "'>='": (70, 3), "'*'": (70, 3), "'}'": (70, 3), "'+'": (70, 3), "'<'": (70, 3)}, {}, ['expr']), ({"'!='": (71, 3), "'>'": (71, 3), 'T.newline': (71, 3), "'='": (71, 3), "'/'": 216, "'-'": (71, 3), "','": (71, 3), "')'": (71, 3), "'<='": (71, 3), "'>='": (71, 3), "'*'": 215, "'}'": (71, 3), "'+'": (71, 3), "'<'": (71, 3)}, {}, ['expr']), ({"'!='": (71, 3), "'>'": (71, 3), 'T.newline': (71, 3), "'='": (71, 3), "'/'": 216, "'-'": (71, 3), "','": (71, 3), "')'": (71, 3), "'<='": (71, 3), "'>='": (71, 3), "'*'": 215, "'}'": (71, 3), "'+'": (71, 3), "'<'": (71, 3)}, {}, ['expr']), ({"'*'": 162, "'/'": 163, "'-'": 161, "'+'": 160, "','": 241}, {}, ['expr', 'max']), ({"'*'": 162, "'/'": 163, "'-'": 161, "'+'": 160, "','": 242}, {}, ['expr', 'min']), ({"'!='": (71, 3), "'>'": (71, 3), 'T.newline': (71, 3), '__end_of_input__': (71, 3), "'='": (71, 3), "'/'": 154, "'-'": (71, 3), "','": (71, 3), "')'": (71, 3), "'<='": (71, 3), "'+'": (71, 3), "'||'": (71, 3), "'>='": (71, 3), "'*'": 153, "'}'": (71, 3), "'&&'": (71, 3), "'<'": (71, 3)}, {}, ['expr']), ({"'!='": (71, 3), "'>'": (71, 3), 'T.newline': (71, 3), '__end_of_input__': (71, 3), "'='": (71, 3), "'/'": 154, "'-'": (71, 3), "','": (71, 3), "')'": (71, 3), "'<='": (71, 3), "'+'": (71, 3), "'||'": (71, 3), "'>='": (71, 3), "'*'": 153, "'}'": (71, 3), "'&&'": (71, 3), "'<'": (71, 3)}, {}, ['expr']), ({"'!='": (70, 3), "'>'": (70, 3), 'T.newline': (70, 3), '__end_of_input__': (70, 3), "'='": (70, 3), "'/'": (70, 3), "'-'": (70, 3), "','": (70, 3), "')'": (70, 3), "'<='": (70, 3), "'+'": (70, 3), "'||'": (70, 3), "'>='": (70, 3), "'*'": (70, 3), "'}'": (70, 3), "'&&'": (70, 3), "'<'": (70, 3)}, {}, ['expr']), ({"'!='": (70, 3), "'>'": (70, 3), 'T.newline': (70, 3), '__end_of_input__': (70, 3), "'='": (70, 3), "'/'": (70, 3), "'-'": (70, 3), "','": (70, 3), "')'": (70, 3), "'<='": (70, 3), "'+'": (70, 3), "'||'": (70, 3), "'>='": (70, 3), "'*'": (70, 3), "'}'": (70, 3), "'&&'": (70, 3), "'<'": (70, 3)}, {}, ['expr']), ({"'{'": 411}, {29: 365}, ['ifs']), ({"')'": (62, 3), "'||'": (62, 3), "'&&'": 221}, {}, ['boolexpr']), ({"')'": (61, 3), "'||'": (61, 3), "'&&'": (61, 3)}, {}, ['boolexpr']), ({"')'": 367, "'*'": 401, "'/'": 402, "'-'": 400, "'+'": 399}, {}, ['expr', 'max']), ({"'min'": 208, 'T.integer': 284, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 115, "'E'": 205, "'+'": 206}, {32: 185, 34: 285, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['max']), ({"'min'": 208, 'T.integer': 284, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 115, "'E'": 205, "'+'": 206}, {32: 185, 34: 286, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['min']), ({"'/'": (69, 2), "'-'": (69, 2), "','": (69, 2), "')'": (69, 2), "'*'": (69, 2), "'+'": (69, 2)}, {}, ['expr']), ({"'min'": 208, 'T.integer': 117, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 184, "'E'": 205, "'+'": 206}, {32: 185, 34: 249, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['max']), ({"'min'": 208, 'T.integer': 117, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 184, "'E'": 205, "'+'": 206}, {32: 185, 34: 250, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['min']), ({"'/'": (69, 2), "'-'": (69, 2), "','": (69, 2), "')'": (69, 2), "'*'": (69, 2), "'+'": (69, 2)}, {}, ['expr']), ({"'/'": (68, 2, 'expt'), "'-'": (68, 2, 'expt'), "','": (68, 2, 'expt'), "')'": (68, 2, 'expt'), "'*'": (68, 2, 'expt'), "'+'": (68, 2, 'expt')}, {}, ['expt']), ({"'/'": (32, 2, 'lval'), "'-'": (32, 2, 'lval'), "','": (32, 2, 'lval'), "')'": (32, 2, 'lval'), "'*'": (32, 2, 'lval'), "'+'": (32, 2, 'lval')}, {}, ['lval']), ({"')'": 447, "'*'": 316, "'/'": 317, "'-'": 315, "'+'": 314}, {}, ['expr', 'expr_p']), ({"')'": 368, "'*'": 401, "'/'": 402, "'-'": 400, "'+'": 399}, {}, ['expr', 'min']), ({"'/'": 317, "'-'": 315, "')'": (63, 3), "'+'": 314, "'||'": (63, 3), "'*'": 316, "'&&'": (63, 3)}, {}, ['boolexpr', 'expr']), ({"')'": 369, "'*'": 401, "'/'": 402, "'-'": 400, "'+'": 399}, {}, ['expr', 'max']), ({"')'": 370, "'*'": 401, "'/'": 402, "'-'": 400, "'+'": 399}, {}, ['expr', 'min']), ({"'!='": (51, 1), "'>'": (51, 1), 'T.newline': (51, 1), '__end_of_input__': (51, 1), "'='": (51, 1), "'/'": (51, 1), "'-'": (51, 1), "','": (51, 1), "')'": (51, 1), "'<='": (51, 1), "'+'": (51, 1), "'||'": (51, 1), "'>='": (51, 1), "'*'": (51, 1), "'}'": (51, 1), "'&&'": (51, 1), "'<'": (51, 1)}, {}, ['expr']), ({"'min'": 208, 'T.integer': 117, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 184, "'E'": 205, "'+'": 206}, {32: 185, 34: 371, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['expr_p']), ({"'!='": (52, 1), "'>'": (52, 1), 'T.newline': (52, 1), '__end_of_input__': (52, 1), "'='": (52, 1), "'/'": (52, 1), "'-'": (52, 1), "','": (52, 1), "')'": (52, 1), "'<='": (52, 1), "'+'": (52, 1), "'||'": (52, 1), "'>='": (52, 1), "'*'": (52, 1), "'}'": (52, 1), "'&&'": (52, 1), "'<'": (52, 1)}, {}, ['expr']), ({"'!='": (24, 1, 'num'), "'>'": (24, 1, 'num'), 'T.newline': (24, 1, 'num'), '__end_of_input__': (24, 1, 'num'), "'='": (24, 1, 'num'), "'/'": (24, 1, 'num'), "'-'": (24, 1, 'num'), "','": (24, 1, 'num'), "')'": (24, 1, 'num'), "'<='": (24, 1, 'num'), "':='": (24, 1, 'num'), "'+'": (24, 1, 'num'), "'||'": (24, 1, 'num'), "'>='": (24, 1, 'num'), "'*'": (24, 1, 'num'), "'}'": (24, 1, 'num'), "'&&'": (24, 1, 'num'), "'<'": (24, 1, 'num')}, {}, ['num']), ({'T.integer': 254}, {24: 372}, ['lval']), ({"'!='": (34, 1, 'expr'), "'>'": (34, 1, 'expr'), 'T.newline': (34, 1, 'expr'), '__end_of_input__': (34, 1, 'expr'), "'='": (34, 1, 'expr'), "'/'": (34, 1, 'expr'), "'-'": (34, 1, 'expr'), "','": (34, 1, 'expr'), "')'": (34, 1, 'expr'), "'<='": (34, 1, 'expr'), "'+'": (34, 1, 'expr'), "'||'": (34, 1, 'expr'), "'>='": (34, 1, 'expr'), "'*'": (34, 1, 'expr'), "'}'": (34, 1, 'expr'), "'&&'": (34, 1, 'expr'), "'<'": (34, 1, 'expr')}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), '__end_of_input__': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'&&'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (53, 1), "'>'": (53, 1), 'T.newline': (53, 1), '__end_of_input__': (53, 1), "'='": (53, 1), "'/'": (53, 1), "'-'": (53, 1), "','": (53, 1), "')'": (53, 1), "'<='": (53, 1), "'+'": (53, 1), "'||'": (53, 1), "'>='": (53, 1), "'*'": (53, 1), "'}'": (53, 1), "'&&'": (53, 1), "'<'": (53, 1)}, {}, ['expr']), ({"'!='": (54, 1), "'>'": (54, 1), 'T.newline': (54, 1), '__end_of_input__': (54, 1), "'='": (54, 1), "'/'": (54, 1), "'-'": (54, 1), "','": (54, 1), "')'": (54, 1), "'<='": (54, 1), "'+'": (54, 1), "'||'": (54, 1), "'>='": (54, 1), "'*'": (54, 1), "'}'": (54, 1), "'&&'": (54, 1), "'<'": (54, 1)}, {}, ['expr']), ({"'!='": (55, 1), "'>'": (55, 1), 'T.newline': (55, 1), '__end_of_input__': (55, 1), "'='": (55, 1), "'/'": (55, 1), "'-'": (55, 1), "','": (55, 1), "')'": (55, 1), "'<='": (55, 1), "'+'": (55, 1), "'||'": (55, 1), "'>='": (55, 1), "'*'": (55, 1), "'}'": (55, 1), "'&&'": (55, 1), "'<'": (55, 1)}, {}, ['expr']), ({"'!='": (56, 1), "'>'": (56, 1), 'T.newline': (56, 1), '__end_of_input__': (56, 1), "'='": (56, 1), "'/'": (56, 1), "'-'": (56, 1), "','": (56, 1), "')'": (56, 1), "'<='": (56, 1), "'+'": (56, 1), "'||'": (56, 1), "'>='": (56, 1), "'*'": (56, 1), "'}'": (56, 1), "'&&'": (56, 1), "'<'": (56, 1)}, {}, ['expr']), ({"'!='": (57, 1), "'>'": (57, 1), 'T.newline': (57, 1), '__end_of_input__': (57, 1), "'='": (57, 1), "'/'": (57, 1), "'-'": (57, 1), "','": (57, 1), "')'": (57, 1), "'<='": (57, 1), "'+'": (57, 1), "'||'": (57, 1), "'>='": (57, 1), "'*'": (57, 1), "'}'": (57, 1), "'&&'": (57, 1), "'<'": (57, 1)}, {}, ['expr']), ({"'!='": (58, 1), "'>'": (58, 1), 'T.newline': (58, 1), '__end_of_input__': (58, 1), "'='": (58, 1), "'/'": (58, 1), "'-'": (58, 1), "','": (58, 1), "')'": (58, 1), "'<='": (58, 1), "'+'": (58, 1), "'||'": (58, 1), "'>='": (58, 1), "'*'": (58, 1), "'}'": (58, 1), "'&&'": (58, 1), "'<'": (58, 1)}, {}, ['expr']), ({"'!='": (59, 1), "'>'": (59, 1), 'T.newline': (59, 1), '__end_of_input__': (59, 1), "'='": (59, 1), "'/'": (59, 1), "'-'": (59, 1), "','": (59, 1), "')'": (59, 1), "'<='": (59, 1), "'+'": (59, 1), "'||'": (59, 1), "'>='": (59, 1), "'*'": (59, 1), "'}'": (59, 1), "'&&'": (59, 1), "'<'": (59, 1)}, {}, ['expr']), ({'T.integer': 254}, {24: 373}, ['expt']), ({"'min'": 276, 'T.integer': 254, "'max'": 277, "'-'": 275, "'V'": 255, "'('": 252, "'E'": 273, "'+'": 274}, {32: 253, 34: 374, 68: 266, 69: 267, 70: 268, 71: 269, 72: 270, 73: 271, 42: 256, 56: 262, 51: 257, 52: 258, 53: 259, 54: 260, 55: 261, 24: 251, 57: 263, 58: 264, 59: 265, 74: 272}, ['expr']), ({"'min'": 276, 'T.integer': 254, "'max'": 277, "'-'": 275, "'V'": 255, "'('": 252, "'E'": 273, "'+'": 274}, {32: 253, 34: 375, 68: 266, 69: 267, 70: 268, 71: 269, 72: 270, 73: 271, 42: 256, 56: 262, 51: 257, 52: 258, 53: 259, 54: 260, 55: 261, 24: 251, 57: 263, 58: 264, 59: 265, 74: 272}, ['expr']), ({"'('": 376}, {}, ['min']), ({"'('": 377}, {}, ['max']), ({"'/'": 317, "'-'": 315, "')'": (63, 3), "'+'": 314, "'||'": (63, 3), "'*'": 316, "'&&'": (63, 3)}, {}, ['boolexpr', 'expr']), ({"'/'": 317, "'-'": 315, "')'": (63, 3), "'+'": 314, "'||'": (63, 3), "'*'": 316, "'&&'": (63, 3)}, {}, ['boolexpr', 'expr']), ({"'/'": 317, "'-'": 315, "')'": (63, 3), "'+'": 314, "'||'": (63, 3), "'*'": 316, "'&&'": (63, 3)}, {}, ['boolexpr', 'expr']), ({"'/'": 317, "'-'": 315, "')'": (63, 3), "'+'": 314, "'||'": (63, 3), "'*'": 316, "'&&'": (63, 3)}, {}, ['boolexpr', 'expr']), ({"'/'": 317, "'-'": 315, "')'": (63, 3), "'+'": 314, "'||'": (63, 3), "'*'": 316, "'&&'": (63, 3)}, {}, ['boolexpr', 'expr']), ({"'*'": 401, "'/'": 402, "'-'": 400, "'+'": 399, "','": 378}, {}, ['expr', 'max']), ({"'*'": (24, 1, 'num'), "'/'": (24, 1, 'num'), "'-'": (24, 1, 'num'), "'+'": (24, 1, 'num'), "','": (24, 1, 'num')}, {}, ['num']), ({"'*'": 316, "'/'": 317, "'-'": 315, "'+'": 314, "','": 445}, {}, ['expr', 'max']), ({"'*'": 316, "'/'": 317, "'-'": 315, "'+'": 314, "','": 446}, {}, ['expr', 'min']), ({"'/'": (51, 1), "'-'": (51, 1), "','": (51, 1), "')'": (51, 1), "'*'": (51, 1), "'+'": (51, 1)}, {}, ['expr']), ({"'min'": 208, 'T.integer': 117, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 184, "'E'": 205, "'+'": 206}, {32: 185, 34: 328, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['expr_p']), ({"'/'": (52, 1), "'-'": (52, 1), "','": (52, 1), "')'": (52, 1), "'*'": (52, 1), "'+'": (52, 1)}, {}, ['expr']), ({"'/'": (24, 1, 'num'), "'-'": (24, 1, 'num'), "','": (24, 1, 'num'), "')'": (24, 1, 'num'), "'*'": (24, 1, 'num'), "'+'": (24, 1, 'num')}, {}, ['num']), ({'T.integer': 290}, {24: 323}, ['lval']), ({"'/'": (34, 1, 'expr'), "'-'": (34, 1, 'expr'), "','": (34, 1, 'expr'), "')'": (34, 1, 'expr'), "'*'": (34, 1, 'expr'), "'+'": (34, 1, 'expr')}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'*'": (42, 1), "'+'": (42, 1)}, {}, ['expr']), ({"'/'": (53, 1), "'-'": (53, 1), "','": (53, 1), "')'": (53, 1), "'*'": (53, 1), "'+'": (53, 1)}, {}, ['expr']), ({"'/'": (54, 1), "'-'": (54, 1), "','": (54, 1), "')'": (54, 1), "'*'": (54, 1), "'+'": (54, 1)}, {}, ['expr']), ({"'/'": (55, 1), "'-'": (55, 1), "','": (55, 1), "')'": (55, 1), "'*'": (55, 1), "'+'": (55, 1)}, {}, ['expr']), ({"'/'": (56, 1), "'-'": (56, 1), "','": (56, 1), "')'": (56, 1), "'*'": (56, 1), "'+'": (56, 1)}, {}, ['expr']), ({"'/'": (57, 1), "'-'": (57, 1), "','": (57, 1), "')'": (57, 1), "'*'": (57, 1), "'+'": (57, 1)}, {}, ['expr']), ({"'/'": (58, 1), "'-'": (58, 1), "','": (58, 1), "')'": (58, 1), "'*'": (58, 1), "'+'": (58, 1)}, {}, ['expr']), ({"'/'": (59, 1), "'-'": (59, 1), "','": (59, 1), "')'": (59, 1), "'*'": (59, 1), "'+'": (59, 1)}, {}, ['expr']), ({'T.integer': 290}, {24: 322}, ['expt']), ({"'min'": 312, 'T.integer': 290, "'max'": 313, "'-'": 311, "'V'": 291, "'('": 288, "'E'": 309, "'+'": 310}, {32: 289, 34: 321, 68: 302, 69: 303, 70: 304, 71: 305, 72: 306, 73: 307, 42: 292, 56: 298, 51: 293, 52: 294, 53: 295, 54: 296, 55: 297, 24: 287, 57: 299, 58: 300, 59: 301, 74: 308}, ['expr']), ({"'min'": 312, 'T.integer': 290, "'max'": 313, "'-'": 311, "'V'": 291, "'('": 288, "'E'": 309, "'+'": 310}, {32: 289, 34: 320, 68: 302, 69: 303, 70: 304, 71: 305, 72: 306, 73: 307, 42: 292, 56: 298, 51: 293, 52: 294, 53: 295, 54: 296, 55: 297, 24: 287, 57: 299, 58: 300, 59: 301, 74: 308}, ['expr']), ({"'('": 319}, {}, ['min']), ({"'('": 318}, {}, ['max']), ({"'min'": 356, 'T.integer': 334, "'max'": 357, "'-'": 355, "'V'": 335, "'('": 332, "'E'": 353, "'+'": 354}, {32: 333, 34: 327, 68: 346, 69: 347, 70: 348, 71: 349, 72: 350, 73: 351, 42: 336, 56: 342, 51: 337, 52: 338, 53: 339, 54: 340, 55: 341, 24: 331, 57: 343, 58: 344, 59: 345, 74: 352}, ['expr']), ({"'min'": 356, 'T.integer': 334, "'max'": 357, "'-'": 355, "'V'": 335, "'('": 332, "'E'": 353, "'+'": 354}, {32: 333, 34: 326, 68: 346, 69: 347, 70: 348, 71: 349, 72: 350, 73: 351, 42: 336, 56: 342, 51: 337, 52: 338, 53: 339, 54: 340, 55: 341, 24: 331, 57: 343, 58: 344, 59: 345, 74: 352}, ['expr']), ({"'min'": 356, 'T.integer': 334, "'max'": 357, "'-'": 355, "'V'": 335, "'('": 332, "'E'": 353, "'+'": 354}, {32: 333, 34: 325, 68: 346, 69: 347, 70: 348, 71: 349, 72: 350, 73: 351, 42: 336, 56: 342, 51: 337, 52: 338, 53: 339, 54: 340, 55: 341, 24: 331, 57: 343, 58: 344, 59: 345, 74: 352}, ['expr']), ({"'min'": 356, 'T.integer': 334, "'max'": 357, "'-'": 355, "'V'": 335, "'('": 332, "'E'": 353, "'+'": 354}, {32: 333, 34: 324, 68: 346, 69: 347, 70: 348, 71: 349, 72: 350, 73: 351, 42: 336, 56: 342, 51: 337, 52: 338, 53: 339, 54: 340, 55: 341, 24: 331, 57: 343, 58: 344, 59: 345, 74: 352}, ['expr']), ({"'min'": 208, 'T.integer': 284, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 115, "'E'": 205, "'+'": 206}, {32: 185, 34: 381, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['max']), ({"'min'": 208, 'T.integer': 284, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 115, "'E'": 205, "'+'": 206}, {32: 185, 34: 382, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['min']), ({"'/'": (69, 2), "'-'": (69, 2), "','": (69, 2), "')'": (69, 2), "'*'": (69, 2), "'+'": (69, 2)}, {}, ['expr']), ({"'/'": (69, 2), "'-'": (69, 2), "','": (69, 2), "')'": (69, 2), "'*'": (69, 2), "'+'": (69, 2)}, {}, ['expr']), ({"'/'": (68, 2, 'expt'), "'-'": (68, 2, 'expt'), "','": (68, 2, 'expt'), "')'": (68, 2, 'expt'), "'*'": (68, 2, 'expt'), "'+'": (68, 2, 'expt')}, {}, ['expt']), ({"'/'": (32, 2, 'lval'), "'-'": (32, 2, 'lval'), "','": (32, 2, 'lval'), "')'": (32, 2, 'lval'), "'*'": (32, 2, 'lval'), "'+'": (32, 2, 'lval')}, {}, ['lval']), ({"'/'": (70, 3), "'-'": (70, 3), "','": (70, 3), "')'": (70, 3), "'+'": (70, 3), "'||'": (70, 3), "'*'": (70, 3), "'&&'": (70, 3)}, {}, ['expr']), ({"'/'": (70, 3), "'-'": (70, 3), "','": (70, 3), "')'": (70, 3), "'+'": (70, 3), "'||'": (70, 3), "'*'": (70, 3), "'&&'": (70, 3)}, {}, ['expr']), ({"'/'": 361, "'-'": (71, 3), "','": (71, 3), "')'": (71, 3), "'+'": (71, 3), "'||'": (71, 3), "'*'": 360, "'&&'": (71, 3)}, {}, ['expr']), ({"'/'": 361, "'-'": (71, 3), "','": (71, 3), "')'": (71, 3), "'+'": (71, 3), "'||'": (71, 3), "'*'": 360, "'&&'": (71, 3)}, {}, ['expr']), ({"')'": 383, "'*'": 316, "'/'": 317, "'-'": 315, "'+'": 314}, {}, ['expr', 'expr_p']), ({"'*'": 401, "'/'": 402, "'-'": 400, "'+'": 399, "','": 384}, {}, ['expr', 'min']), ({"'!='": (72, 3, 'expr_p'), "'>'": (72, 3, 'expr_p'), 'T.newline': (72, 3, 'expr_p'), "'='": (72, 3, 'expr_p'), "'/'": (72, 3, 'expr_p'), "'-'": (72, 3, 'expr_p'), "','": (72, 3, 'expr_p'), "')'": (72, 3, 'expr_p'), "'<='": (72, 3, 'expr_p'), "'>='": (72, 3, 'expr_p'), "'*'": (72, 3, 'expr_p'), "'}'": (72, 3, 'expr_p'), "'+'": (72, 3, 'expr_p'), "'<'": (72, 3, 'expr_p')}, {}, ['expr_p']), ({"'/'": (51, 1), "'-'": (51, 1), "','": (51, 1), "')'": (51, 1), "'+'": (51, 1), "'||'": (51, 1), "'*'": (51, 1), "'&&'": (51, 1)}, {}, ['expr']), ({"'min'": 208, 'T.integer': 117, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 184, "'E'": 205, "'+'": 206}, {32: 185, 34: 385, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['expr_p']), ({"'/'": (52, 1), "'-'": (52, 1), "','": (52, 1), "')'": (52, 1), "'+'": (52, 1), "'||'": (52, 1), "'*'": (52, 1), "'&&'": (52, 1)}, {}, ['expr']), ({"'/'": (24, 1, 'num'), "'-'": (24, 1, 'num'), "','": (24, 1, 'num'), "')'": (24, 1, 'num'), "'+'": (24, 1, 'num'), "'||'": (24, 1, 'num'), "'*'": (24, 1, 'num'), "'&&'": (24, 1, 'num')}, {}, ['num']), ({'T.integer': 334}, {24: 386}, ['lval']), ({"'/'": (34, 1, 'expr'), "'-'": (34, 1, 'expr'), "','": (34, 1, 'expr'), "')'": (34, 1, 'expr'), "'+'": (34, 1, 'expr'), "'||'": (34, 1, 'expr'), "'*'": (34, 1, 'expr'), "'&&'": (34, 1, 'expr')}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'*'": (42, 1), "'&&'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'*'": (42, 1), "'&&'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'*'": (42, 1), "'&&'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'*'": (42, 1), "'&&'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'*'": (42, 1), "'&&'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'*'": (42, 1), "'&&'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'*'": (42, 1), "'&&'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'*'": (42, 1), "'&&'": (42, 1)}, {}, ['expr']), ({"'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'+'": (42, 1), "'||'": (42, 1), "'*'": (42, 1), "'&&'": (42, 1)}, {}, ['expr']), ({"'/'": (53, 1), "'-'": (53, 1), "','": (53, 1), "')'": (53, 1), "'+'": (53, 1), "'||'": (53, 1), "'*'": (53, 1), "'&&'": (53, 1)}, {}, ['expr']), ({"'/'": (54, 1), "'-'": (54, 1), "','": (54, 1), "')'": (54, 1), "'+'": (54, 1), "'||'": (54, 1), "'*'": (54, 1), "'&&'": (54, 1)}, {}, ['expr']), ({"'/'": (55, 1), "'-'": (55, 1), "','": (55, 1), "')'": (55, 1), "'+'": (55, 1), "'||'": (55, 1), "'*'": (55, 1), "'&&'": (55, 1)}, {}, ['expr']), ({"'/'": (56, 1), "'-'": (56, 1), "','": (56, 1), "')'": (56, 1), "'+'": (56, 1), "'||'": (56, 1), "'*'": (56, 1), "'&&'": (56, 1)}, {}, ['expr']), ({"'/'": (57, 1), "'-'": (57, 1), "','": (57, 1), "')'": (57, 1), "'+'": (57, 1), "'||'": (57, 1), "'*'": (57, 1), "'&&'": (57, 1)}, {}, ['expr']), ({"'/'": (58, 1), "'-'": (58, 1), "','": (58, 1), "')'": (58, 1), "'+'": (58, 1), "'||'": (58, 1), "'*'": (58, 1), "'&&'": (58, 1)}, {}, ['expr']), ({"'/'": (59, 1), "'-'": (59, 1), "','": (59, 1), "')'": (59, 1), "'+'": (59, 1), "'||'": (59, 1), "'*'": (59, 1), "'&&'": (59, 1)}, {}, ['expr']), ({'T.integer': 334}, {24: 387}, ['expt']), ({"'min'": 356, 'T.integer': 334, "'max'": 357, "'-'": 355, "'V'": 335, "'('": 332, "'E'": 353, "'+'": 354}, {32: 333, 34: 388, 68: 346, 69: 347, 70: 348, 71: 349, 72: 350, 73: 351, 42: 336, 56: 342, 51: 337, 52: 338, 53: 339, 54: 340, 55: 341, 24: 331, 57: 343, 58: 344, 59: 345, 74: 352}, ['expr']), ({"'min'": 356, 'T.integer': 334, "'max'": 357, "'-'": 355, "'V'": 335, "'('": 332, "'E'": 353, "'+'": 354}, {32: 333, 34: 389, 68: 346, 69: 347, 70: 348, 71: 349, 72: 350, 73: 351, 42: 336, 56: 342, 51: 337, 52: 338, 53: 339, 54: 340, 55: 341, 24: 331, 57: 343, 58: 344, 59: 345, 74: 352}, ['expr']), ({"'('": 390}, {}, ['min']), ({"'('": 391}, {}, ['max']), ({"'min'": 356, 'T.integer': 334, "'max'": 357, "'-'": 355, "'V'": 335, "'('": 332, "'E'": 353, "'+'": 354}, {32: 333, 34: 392, 68: 346, 69: 347, 70: 348, 71: 349, 72: 350, 73: 351, 42: 336, 56: 342, 51: 337, 52: 338, 53: 339, 54: 340, 55: 341, 24: 331, 57: 343, 58: 344, 59: 345, 74: 352}, ['expr']), ({"'min'": 356, 'T.integer': 334, "'max'": 357, "'-'": 355, "'V'": 335, "'('": 332, "'E'": 353, "'+'": 354}, {32: 333, 34: 393, 68: 346, 69: 347, 70: 348, 71: 349, 72: 350, 73: 351, 42: 336, 56: 342, 51: 337, 52: 338, 53: 339, 54: 340, 55: 341, 24: 331, 57: 343, 58: 344, 59: 345, 74: 352}, ['expr']), ({"'min'": 356, 'T.integer': 334, "'max'": 357, "'-'": 355, "'V'": 335, "'('": 332, "'E'": 353, "'+'": 354}, {32: 333, 34: 394, 68: 346, 69: 347, 70: 348, 71: 349, 72: 350, 73: 351, 42: 336, 56: 342, 51: 337, 52: 338, 53: 339, 54: 340, 55: 341, 24: 331, 57: 343, 58: 344, 59: 345, 74: 352}, ['expr']), ({"'min'": 356, 'T.integer': 334, "'max'": 357, "'-'": 355, "'V'": 335, "'('": 332, "'E'": 353, "'+'": 354}, {32: 333, 34: 395, 68: 346, 69: 347, 70: 348, 71: 349, 72: 350, 73: 351, 42: 336, 56: 342, 51: 337, 52: 338, 53: 339, 54: 340, 55: 341, 24: 331, 57: 343, 58: 344, 59: 345, 74: 352}, ['expr']), ({"')'": (67, 3, 'boolexpr_p'), "'||'": (67, 3, 'boolexpr_p'), "'&&'": (67, 3, 'boolexpr_p')}, {}, ['boolexpr_p']), ({'__end_of_input__': (20, 5, 'whiles'), "'}'": (20, 5, 'whiles'), 'T.newline': (20, 5, 'whiles')}, {}, ['whiles']), ({'T.newline': 366}, {}, ['then']), ({'__end_of_input__': (30, 0, 'elses'), "'else'": 397, "'}'": (30, 0, 'elses'), 'T.newline': (30, 0, 'elses')}, {30: 396}, ['ifs']), ({"'S'": 20, "'skip'": 24, "'if'": 21, "'while'": 22, "'V'": 25}, {32: 23, 4: 410, 5: 3, 6: 4, 7: 5, 8: 6, 9: 7, 11: 8, 12: 9, 13: 10, 14: 11, 15: 12, 16: 13, 17: 14, 18: 15, 19: 16, 20: 17, 21: 18, 22: 19}, ['then']), ({"'!='": (74, 6, 'max'), "'>'": (74, 6, 'max'), 'T.newline': (74, 6, 'max'), '__end_of_input__': (74, 6, 'max'), "'='": (74, 6, 'max'), "'/'": (74, 6, 'max'), "'-'": (74, 6, 'max'), "','": (74, 6, 'max'), "')'": (74, 6, 'max'), "'<='": (74, 6, 'max'), "'>='": (74, 6, 'max'), "'*'": (74, 6, 'max'), "'}'": (74, 6, 'max'), "'+'": (74, 6, 'max'), "'<'": (74, 6, 'max')}, {}, ['max']), ({"'!='": (73, 6, 'min'), "'>'": (73, 6, 'min'), 'T.newline': (73, 6, 'min'), '__end_of_input__': (73, 6, 'min'), "'='": (73, 6, 'min'), "'/'": (73, 6, 'min'), "'-'": (73, 6, 'min'), "','": (73, 6, 'min'), "')'": (73, 6, 'min'), "'<='": (73, 6, 'min'), "'>='": (73, 6, 'min'), "'*'": (73, 6, 'min'), "'}'": (73, 6, 'min'), "'+'": (73, 6, 'min'), "'<'": (73, 6, 'min')}, {}, ['min']), ({"'!='": (74, 6, 'max'), "'>'": (74, 6, 'max'), 'T.newline': (74, 6, 'max'), '__end_of_input__': (74, 6, 'max'), "'/'": (74, 6, 'max'), "'-'": (74, 6, 'max'), "','": (74, 6, 'max'), "')'": (74, 6, 'max'), "'<='": (74, 6, 'max'), "'+'": (74, 6, 'max'), "'||'": (74, 6, 'max'), "'>='": (74, 6, 'max'), "'*'": (74, 6, 'max'), "'='": (74, 6, 'max'), "'&&'": (74, 6, 'max'), "'<'": (74, 6, 'max')}, {}, ['max']), ({"'!='": (73, 6, 'min'), "'>'": (73, 6, 'min'), 'T.newline': (73, 6, 'min'), '__end_of_input__': (73, 6, 'min'), "'/'": (73, 6, 'min'), "'-'": (73, 6, 'min'), "','": (73, 6, 'min'), "')'": (73, 6, 'min'), "'<='": (73, 6, 'min'), "'+'": (73, 6, 'min'), "'||'": (73, 6, 'min'), "'>='": (73, 6, 'min'), "'*'": (73, 6, 'min'), "'='": (73, 6, 'min'), "'&&'": (73, 6, 'min'), "'<'": (73, 6, 'min')}, {}, ['min']), ({"')'": 409, "'*'": 401, "'/'": 402, "'-'": 400, "'+'": 399}, {}, ['expr', 'expr_p']), ({"'!='": (32, 2, 'lval'), "'>'": (32, 2, 'lval'), 'T.newline': (32, 2, 'lval'), '__end_of_input__': (32, 2, 'lval'), "'='": (32, 2, 'lval'), "'/'": (32, 2, 'lval'), "'-'": (32, 2, 'lval'), "','": (32, 2, 'lval'), "')'": (32, 2, 'lval'), "'<='": (32, 2, 'lval'), "'+'": (32, 2, 'lval'), "'||'": (32, 2, 'lval'), "'>='": (32, 2, 'lval'), "'*'": (32, 2, 'lval'), "'}'": (32, 2, 'lval'), "'&&'": (32, 2, 'lval'), "'<'": (32, 2, 'lval')}, {}, ['lval']), ({"'!='": (68, 2, 'expt'), "'>'": (68, 2, 'expt'), 'T.newline': (68, 2, 'expt'), '__end_of_input__': (68, 2, 'expt'), "'='": (68, 2, 'expt'), "'/'": (68, 2, 'expt'), "'-'": (68, 2, 'expt'), "','": (68, 2, 'expt'), "')'": (68, 2, 'expt'), "'<='": (68, 2, 'expt'), "'+'": (68, 2, 'expt'), "'||'": (68, 2, 'expt'), "'>='": (68, 2, 'expt'), "'*'": (68, 2, 'expt'), "'}'": (68, 2, 'expt'), "'&&'": (68, 2, 'expt'), "'<'": (68, 2, 'expt')}, {}, ['expt']), ({"'!='": (69, 2), "'>'": (69, 2), 'T.newline': (69, 2), '__end_of_input__': (69, 2), "'='": (69, 2), "'/'": (69, 2), "'-'": (69, 2), "','": (69, 2), "')'": (69, 2), "'<='": (69, 2), "'+'": (69, 2), "'||'": (69, 2), "'>='": (69, 2), "'*'": (69, 2), "'}'": (69, 2), "'&&'": (69, 2), "'<'": (69, 2)}, {}, ['expr']), ({"'!='": (69, 2), "'>'": (69, 2), 'T.newline': (69, 2), '__end_of_input__': (69, 2), "'='": (69, 2), "'/'": (69, 2), "'-'": (69, 2), "','": (69, 2), "')'": (69, 2), "'<='": (69, 2), "'+'": (69, 2), "'||'": (69, 2), "'>='": (69, 2), "'*'": (69, 2), "'}'": (69, 2), "'&&'": (69, 2), "'<'": (69, 2)}, {}, ['expr']), ({"'min'": 208, 'T.integer': 284, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 115, "'E'": 205, "'+'": 206}, {32: 185, 34: 408, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['min']), ({"'min'": 208, 'T.integer': 284, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 115, "'E'": 205, "'+'": 206}, {32: 185, 34: 407, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['max']), ({"'min'": 208, 'T.integer': 117, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 184, "'E'": 205, "'+'": 206}, {32: 185, 34: 413, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['max']), ({"'min'": 112, 'T.integer': 39, "'max'": 113, "'-'": 111, "'V'": 91, "'('": 36, "'E'": 109, "'+'": 110}, {32: 89, 34: 414, 68: 102, 69: 103, 70: 104, 71: 105, 72: 106, 73: 107, 42: 92, 56: 98, 51: 93, 52: 94, 53: 95, 54: 96, 55: 97, 24: 88, 57: 99, 58: 100, 59: 101, 74: 108}, ['max']), ({"'min'": 112, 'T.integer': 39, "'max'": 113, "'-'": 111, "'V'": 91, "'('": 36, "'E'": 109, "'+'": 110}, {32: 89, 34: 415, 68: 102, 69: 103, 70: 104, 71: 105, 72: 106, 73: 107, 42: 92, 56: 98, 51: 93, 52: 94, 53: 95, 54: 96, 55: 97, 24: 88, 57: 99, 58: 100, 59: 101, 74: 108}, ['min']), ({"'*'": 401, "'/'": 402, "'-'": 400, "'+'": 399, "','": 379}, {}, ['expr', 'max']), ({"'*'": 401, "'/'": 402, "'-'": 400, "'+'": 399, "','": 380}, {}, ['expr', 'min']), ({"'/'": (72, 3, 'expr_p'), "'-'": (72, 3, 'expr_p'), "','": (72, 3, 'expr_p'), "')'": (72, 3, 'expr_p'), "'+'": (72, 3, 'expr_p'), "'||'": (72, 3, 'expr_p'), "'*'": (72, 3, 'expr_p'), "'&&'": (72, 3, 'expr_p')}, {}, ['expr_p']), ({"'min'": 208, 'T.integer': 117, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 184, "'E'": 205, "'+'": 206}, {32: 185, 34: 418, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['min']), ({"')'": 383, "'*'": 316, "'/'": 317, "'-'": 315, "'+'": 314}, {}, ['expr', 'expr_p']), ({"'/'": (32, 2, 'lval'), "'-'": (32, 2, 'lval'), "','": (32, 2, 'lval'), "')'": (32, 2, 'lval'), "'+'": (32, 2, 'lval'), "'||'": (32, 2, 'lval'), "'*'": (32, 2, 'lval'), "'&&'": (32, 2, 'lval')}, {}, ['lval']), ({"'/'": (68, 2, 'expt'), "'-'": (68, 2, 'expt'), "','": (68, 2, 'expt'), "')'": (68, 2, 'expt'), "'+'": (68, 2, 'expt'), "'||'": (68, 2, 'expt'), "'*'": (68, 2, 'expt'), "'&&'": (68, 2, 'expt')}, {}, ['expt']), ({"'/'": (69, 2), "'-'": (69, 2), "','": (69, 2), "')'": (69, 2), "'+'": (69, 2), "'||'": (69, 2), "'*'": (69, 2), "'&&'": (69, 2)}, {}, ['expr']), ({"'/'": (69, 2), "'-'": (69, 2), "','": (69, 2), "')'": (69, 2), "'+'": (69, 2), "'||'": (69, 2), "'*'": (69, 2), "'&&'": (69, 2)}, {}, ['expr']), ({"'min'": 208, 'T.integer': 284, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 115, "'E'": 205, "'+'": 206}, {32: 185, 34: 382, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['min']), ({"'min'": 208, 'T.integer': 284, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 115, "'E'": 205, "'+'": 206}, {32: 185, 34: 381, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['max']), ({"'/'": 361, "'-'": (71, 3), "','": (71, 3), "')'": (71, 3), "'+'": (71, 3), "'||'": (71, 3), "'*'": 360, "'&&'": (71, 3)}, {}, ['expr']), ({"'/'": 361, "'-'": (71, 3), "','": (71, 3), "')'": (71, 3), "'+'": (71, 3), "'||'": (71, 3), "'*'": 360, "'&&'": (71, 3)}, {}, ['expr']), ({"'/'": (70, 3), "'-'": (70, 3), "','": (70, 3), "')'": (70, 3), "'+'": (70, 3), "'||'": (70, 3), "'*'": (70, 3), "'&&'": (70, 3)}, {}, ['expr']), ({"'/'": (70, 3), "'-'": (70, 3), "','": (70, 3), "')'": (70, 3), "'+'": (70, 3), "'||'": (70, 3), "'*'": (70, 3), "'&&'": (70, 3)}, {}, ['expr']), ({'__end_of_input__': (19, 6, 'ifs'), "'}'": (19, 6, 'ifs'), 'T.newline': (19, 6, 'ifs')}, {}, ['ifs']), ({"'{'": 411}, {29: 412}, ['elses']), ({"'}'": 420, 'T.newline': 419}, {}, ['stmts', 'stmts2', 'then']), ({"'min'": 312, 'T.integer': 290, "'max'": 313, "'-'": 311, "'V'": 291, "'('": 288, "'E'": 309, "'+'": 310}, {32: 289, 34: 406, 68: 302, 69: 303, 70: 304, 71: 305, 72: 306, 73: 307, 42: 292, 56: 298, 51: 293, 52: 294, 53: 295, 54: 296, 55: 297, 24: 287, 57: 299, 58: 300, 59: 301, 74: 308}, ['expr']), ({"'min'": 312, 'T.integer': 290, "'max'": 313, "'-'": 311, "'V'": 291, "'('": 288, "'E'": 309, "'+'": 310}, {32: 289, 34: 405, 68: 302, 69: 303, 70: 304, 71: 305, 72: 306, 73: 307, 42: 292, 56: 298, 51: 293, 52: 294, 53: 295, 54: 296, 55: 297, 24: 287, 57: 299, 58: 300, 59: 301, 74: 308}, ['expr']), ({"'min'": 312, 'T.integer': 290, "'max'": 313, "'-'": 311, "'V'": 291, "'('": 288, "'E'": 309, "'+'": 310}, {32: 289, 34: 404, 68: 302, 69: 303, 70: 304, 71: 305, 72: 306, 73: 307, 42: 292, 56: 298, 51: 293, 52: 294, 53: 295, 54: 296, 55: 297, 24: 287, 57: 299, 58: 300, 59: 301, 74: 308}, ['expr']), ({"'min'": 312, 'T.integer': 290, "'max'": 313, "'-'": 311, "'V'": 291, "'('": 288, "'E'": 309, "'+'": 310}, {32: 289, 34: 403, 68: 302, 69: 303, 70: 304, 71: 305, 72: 306, 73: 307, 42: 292, 56: 298, 51: 293, 52: 294, 53: 295, 54: 296, 55: 297, 24: 287, 57: 299, 58: 300, 59: 301, 74: 308}, ['expr']), ({"'/'": (70, 3), "'-'": (70, 3), "','": (70, 3), "')'": (70, 3), "'*'": (70, 3), "'+'": (70, 3)}, {}, ['expr']), ({"'/'": (70, 3), "'-'": (70, 3), "','": (70, 3), "')'": (70, 3), "'*'": (70, 3), "'+'": (70, 3)}, {}, ['expr']), ({"'/'": 317, "'-'": (71, 3), "','": (71, 3), "')'": (71, 3), "'*'": 316, "'+'": (71, 3)}, {}, ['expr']), ({"'/'": 317, "'-'": (71, 3), "','": (71, 3), "')'": (71, 3), "'*'": 316, "'+'": (71, 3)}, {}, ['expr']), ({"'*'": 316, "'/'": 317, "'-'": 315, "'+'": 314, "','": 421}, {}, ['expr', 'max']), ({"'*'": 316, "'/'": 317, "'-'": 315, "'+'": 314, "','": 422}, {}, ['expr', 'min']), ({"'!='": (72, 3, 'expr_p'), "'>'": (72, 3, 'expr_p'), 'T.newline': (72, 3, 'expr_p'), '__end_of_input__': (72, 3, 'expr_p'), "'='": (72, 3, 'expr_p'), "'/'": (72, 3, 'expr_p'), "'-'": (72, 3, 'expr_p'), "','": (72, 3, 'expr_p'), "')'": (72, 3, 'expr_p'), "'<='": (72, 3, 'expr_p'), "'+'": (72, 3, 'expr_p'), "'||'": (72, 3, 'expr_p'), "'>='": (72, 3, 'expr_p'), "'*'": (72, 3, 'expr_p'), "'}'": (72, 3, 'expr_p'), "'&&'": (72, 3, 'expr_p'), "'<'": (72, 3, 'expr_p')}, {}, ['expr_p']), ({"'}'": 420, 'T.newline': 423}, {}, ['stmts', 'stmts2', 'then']), ({'T.newline': 424}, {}, ['then']), ({'__end_of_input__': (30, 2, 'elses'), "'}'": (30, 2, 'elses'), 'T.newline': (30, 2, 'elses')}, {}, ['elses']), ({"')'": 425, "'*'": 401, "'/'": 402, "'-'": 400, "'+'": 399}, {}, ['expr', 'max']), ({"')'": 416, "'*'": 316, "'/'": 317, "'-'": 315, "'+'": 314}, {}, ['expr', 'max']), ({"')'": 417, "'*'": 316, "'/'": 317, "'-'": 315, "'+'": 314}, {}, ['expr', 'min']), ({"'/'": (74, 6, 'max'), "'-'": (74, 6, 'max'), "','": (74, 6, 'max'), "')'": (74, 6, 'max'), "'+'": (74, 6, 'max'), "'||'": (74, 6, 'max'), "'*'": (74, 6, 'max'), "'&&'": (74, 6, 'max')}, {}, ['max']), ({"'/'": (73, 6, 'min'), "'-'": (73, 6, 'min'), "','": (73, 6, 'min'), "')'": (73, 6, 'min'), "'+'": (73, 6, 'min'), "'||'": (73, 6, 'min'), "'*'": (73, 6, 'min'), "'&&'": (73, 6, 'min')}, {}, ['min']), ({"')'": 426, "'*'": 401, "'/'": 402, "'-'": 400, "'+'": 399}, {}, ['expr', 'min']), ({"'skip'": 442, 'T.newline': (7, 2), "'V'": 25, "'if'": 439, "'S'": 438, "'}'": 443, "'while'": 440}, {32: 441, 9: 34, 12: 427, 13: 428, 14: 429, 15: 430, 16: 431, 17: 432, 18: 433, 19: 434, 20: 435, 21: 436, 22: 437}, ['stmts', 'stmts2', 'then']), ({'__end_of_input__': (29, 4, 'then'), "'else'": (29, 4, 'then'), "'}'": (29, 4, 'then'), 'T.newline': (29, 4, 'then')}, {}, ['then']), ({"'min'": 208, 'T.integer': 117, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 184, "'E'": 205, "'+'": 206}, {32: 185, 34: 444, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['max']), ({"'min'": 208, 'T.integer': 117, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 184, "'E'": 205, "'+'": 206}, {32: 185, 34: 448, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['min']), ({"'skip'": 442, 'T.newline': (7, 2), "'V'": 25, "'if'": 439, "'S'": 438, "'}'": 443, "'while'": 440}, {32: 441, 9: 34, 12: 427, 13: 428, 14: 429, 15: 430, 16: 431, 17: 432, 18: 433, 19: 434, 20: 435, 21: 436, 22: 437}, ['stmts', 'stmts2', 'then']), ({"'S'": 20, "'skip'": 24, "'if'": 21, "'while'": 22, "'V'": 25}, {32: 23, 4: 449, 5: 3, 6: 4, 7: 5, 8: 6, 9: 7, 11: 8, 12: 9, 13: 10, 14: 11, 15: 12, 16: 13, 17: 14, 18: 15, 19: 16, 20: 17, 21: 18, 22: 19}, ['then']), ({"'!='": (74, 6, 'max'), "'>'": (74, 6, 'max'), 'T.newline': (74, 6, 'max'), "'='": (74, 6, 'max'), "'/'": (74, 6, 'max'), "'-'": (74, 6, 'max'), "','": (74, 6, 'max'), "')'": (74, 6, 'max'), "'<='": (74, 6, 'max'), "'>='": (74, 6, 'max'), "'*'": (74, 6, 'max'), "'}'": (74, 6, 'max'), "'+'": (74, 6, 'max'), "'<'": (74, 6, 'max')}, {}, ['max']), ({"'!='": (73, 6, 'min'), "'>'": (73, 6, 'min'), 'T.newline': (73, 6, 'min'), "'='": (73, 6, 'min'), "'/'": (73, 6, 'min'), "'-'": (73, 6, 'min'), "','": (73, 6, 'min'), "')'": (73, 6, 'min'), "'<='": (73, 6, 'min'), "'>='": (73, 6, 'min'), "'*'": (73, 6, 'min'), "'}'": (73, 6, 'min'), "'+'": (73, 6, 'min'), "'<'": (73, 6, 'min')}, {}, ['min']), ({"'}'": (9, 1, 'stmt'), 'T.newline': (9, 1, 'stmt')}, {}, ['stmt']), ({"'}'": (12, 1), 'T.newline': (12, 1)}, {}, ['stmt']), ({"'}'": (12, 1), 'T.newline': (12, 1)}, {}, ['stmt']), ({"'}'": (12, 1), 'T.newline': (12, 1)}, {}, ['stmt']), ({"'}'": (12, 1), 'T.newline': (12, 1)}, {}, ['stmt']), ({"'}'": (12, 1), 'T.newline': (12, 1)}, {}, ['stmt']), ({"'}'": (13, 1), 'T.newline': (13, 1)}, {}, ['stmt']), ({"'}'": (14, 1), 'T.newline': (14, 1)}, {}, ['stmt']), ({"'}'": (15, 1), 'T.newline': (15, 1)}, {}, ['stmt']), ({"'}'": (16, 1), 'T.newline': (16, 1)}, {}, ['stmt']), ({"'}'": (17, 1), 'T.newline': (17, 1)}, {}, ['stmt']), ({'T.integer': 39}, {24: 450}, ['stmtt']), ({"'('": 451}, {}, ['ifs']), ({"'('": 452}, {}, ['whiles']), ({"':='": 453}, {}, ['assgn']), ({"'}'": (22, 1, 'skip'), 'T.newline': (22, 1, 'skip')}, {}, ['skip']), ({'__end_of_input__': (29, 5, 'then'), "'else'": (29, 5, 'then'), "'}'": (29, 5, 'then'), 'T.newline': (29, 5, 'then')}, {}, ['then']), ({"')'": 454, "'*'": 316, "'/'": 317, "'-'": 315, "'+'": 314}, {}, ['expr', 'max']), ({"'min'": 112, 'T.integer': 117, "'max'": 113, "'-'": 111, "'V'": 91, "'('": 184, "'E'": 109, "'+'": 110}, {32: 89, 34: 414, 68: 102, 69: 103, 70: 104, 71: 105, 72: 106, 73: 107, 42: 92, 56: 98, 51: 93, 52: 94, 53: 95, 54: 96, 55: 97, 24: 88, 57: 99, 58: 100, 59: 101, 74: 108}, ['max']), ({"'min'": 112, 'T.integer': 117, "'max'": 113, "'-'": 111, "'V'": 91, "'('": 184, "'E'": 109, "'+'": 110}, {32: 89, 34: 415, 68: 102, 69: 103, 70: 104, 71: 105, 72: 106, 73: 107, 42: 92, 56: 98, 51: 93, 52: 94, 53: 95, 54: 96, 55: 97, 24: 88, 57: 99, 58: 100, 59: 101, 74: 108}, ['min']), ({"'/'": (72, 3, 'expr_p'), "'-'": (72, 3, 'expr_p'), "','": (72, 3, 'expr_p'), "')'": (72, 3, 'expr_p'), "'*'": (72, 3, 'expr_p'), "'+'": (72, 3, 'expr_p')}, {}, ['expr_p']), ({"')'": 456, "'*'": 316, "'/'": 317, "'-'": 315, "'+'": 314}, {}, ['expr', 'min']), ({"'}'": 458, 'T.newline': 457}, {}, ['stmts', 'stmts2', 'then']), ({"'}'": (18, 2, 'stmtt'), 'T.newline': (18, 2, 'stmtt')}, {}, ['stmtt']), ({"'B'": 85, "'true'": 83, "'min'": 60, "'false'": 84, 'T.integer': 29, "'max'": 61, "'-'": 59, "'V'": 25, "'('": 62, "'E'": 57, "'!'": 82, "'+'": 58}, {24: 35, 27: 459, 32: 37, 34: 64, 37: 65, 42: 40, 43: 66, 44: 67, 45: 68, 46: 69, 47: 70, 48: 71, 49: 72, 50: 73, 51: 41, 52: 42, 53: 43, 54: 44, 55: 45, 56: 46, 57: 47, 58: 48, 59: 49, 60: 74, 61: 75, 62: 76, 63: 77, 64: 78, 65: 79, 66: 80, 67: 81, 68: 50, 69: 51, 70: 52, 71: 53, 72: 54, 73: 55, 74: 56}, ['ifs']), ({"'B'": 85, "'true'": 83, "'min'": 60, "'false'": 84, 'T.integer': 29, "'max'": 61, "'-'": 59, "'V'": 25, "'('": 62, "'E'": 57, "'!'": 82, "'+'": 58}, {24: 35, 27: 461, 32: 37, 34: 64, 37: 65, 42: 40, 43: 66, 44: 67, 45: 68, 46: 69, 47: 70, 48: 71, 49: 72, 50: 73, 51: 41, 52: 42, 53: 43, 54: 44, 55: 45, 56: 46, 57: 47, 58: 48, 59: 49, 60: 74, 61: 75, 62: 76, 63: 77, 64: 78, 65: 79, 66: 80, 67: 81, 68: 50, 69: 51, 70: 52, 71: 53, 72: 54, 73: 55, 74: 56}, ['whiles']), ({"'min'": 143, 'T.integer': 29, "'max'": 144, "'-'": 142, "'V'": 122, "'('": 115, "'E'": 140, "'+'": 141}, {32: 120, 34: 462, 68: 133, 69: 134, 70: 135, 71: 136, 72: 137, 73: 138, 42: 123, 56: 129, 51: 124, 52: 125, 53: 126, 54: 127, 55: 128, 24: 119, 57: 130, 58: 131, 59: 132, 74: 139}, ['assgn']), ({"'!='": (74, 6, 'max'), "'>'": (74, 6, 'max'), 'T.newline': (74, 6, 'max'), '__end_of_input__': (74, 6, 'max'), "'='": (74, 6, 'max'), "'/'": (74, 6, 'max'), "'-'": (74, 6, 'max'), "','": (74, 6, 'max'), "')'": (74, 6, 'max'), "'<='": (74, 6, 'max'), "'+'": (74, 6, 'max'), "'||'": (74, 6, 'max'), "'>='": (74, 6, 'max'), "'*'": (74, 6, 'max'), "'}'": (74, 6, 'max'), "'&&'": (74, 6, 'max'), "'<'": (74, 6, 'max')}, {}, ['max']), ({"'!='": (32, 2, 'lval'), "'>'": (32, 2, 'lval'), 'T.newline': (32, 2, 'lval'), '__end_of_input__': (32, 2, 'lval'), "'/'": (32, 2, 'lval'), "'-'": (32, 2, 'lval'), "','": (32, 2, 'lval'), "')'": (32, 2, 'lval'), "'<='": (32, 2, 'lval'), "'+'": (32, 2, 'lval'), "'||'": (32, 2, 'lval'), "'>='": (32, 2, 'lval'), "'*'": (32, 2, 'lval'), "'='": (32, 2, 'lval'), "'&&'": (32, 2, 'lval'), "'<'": (32, 2, 'lval')}, {}, ['lval']), ({"'!='": (73, 6, 'min'), "'>'": (73, 6, 'min'), 'T.newline': (73, 6, 'min'), '__end_of_input__': (73, 6, 'min'), "'='": (73, 6, 'min'), "'/'": (73, 6, 'min'), "'-'": (73, 6, 'min'), "','": (73, 6, 'min'), "')'": (73, 6, 'min'), "'<='": (73, 6, 'min'), "'+'": (73, 6, 'min'), "'||'": (73, 6, 'min'), "'>='": (73, 6, 'min'), "'*'": (73, 6, 'min'), "'}'": (73, 6, 'min'), "'&&'": (73, 6, 'min'), "'<'": (73, 6, 'min')}, {}, ['min']), ({"'skip'": 442, 'T.newline': (7, 2), "'V'": 25, "'if'": 439, "'S'": 438, "'}'": 488, "'while'": 440}, {32: 441, 9: 34, 12: 427, 13: 428, 14: 429, 15: 430, 16: 431, 17: 432, 18: 433, 19: 434, 20: 435, 21: 436, 22: 437}, ['stmts', 'stmts2', 'then']), ({'__end_of_input__': (29, 4, 'then'), "'else'": (29, 4, 'then'), "'}'": (29, 4, 'then'), 'T.newline': (29, 4, 'then')}, {}, ['then']), ({"')'": 489, "'||'": 222, "'&&'": 221}, {}, ['boolexpr', 'ifs']), ({"'!='": 173, "'>'": 176, "'/'": 163, "'-'": 161, "')'": 223, "'<='": 175, "'>='": 177, "'*'": 162, "'='": 172, "'+'": 160, "'<'": 174}, {}, ['boolexpr', 'expr', 'expr_p']), ({"')'": 490, "'||'": 222, "'&&'": 221}, {}, ['boolexpr', 'whiles']), ({'T.newline': (21, 3, 'assgn'), "'/'": 163, "'-'": 161, "'*'": 162, "'}'": (21, 3, 'assgn'), "'+'": 160}, {}, ['assgn', 'expr']), ({"'!='": (51, 1), "'>'": (51, 1), 'T.newline': (51, 1), "'='": (51, 1), "'/'": (51, 1), "'-'": (51, 1), "','": (51, 1), "')'": (51, 1), "'<='": (51, 1), "'>='": (51, 1), "'*'": (51, 1), "'}'": (51, 1), "'+'": (51, 1), "'<'": (51, 1)}, {}, ['expr']), ({"'!='": (52, 1), "'>'": (52, 1), 'T.newline': (52, 1), "'='": (52, 1), "'/'": (52, 1), "'-'": (52, 1), "','": (52, 1), "')'": (52, 1), "'<='": (52, 1), "'>='": (52, 1), "'*'": (52, 1), "'}'": (52, 1), "'+'": (52, 1), "'<'": (52, 1)}, {}, ['expr']), ({'T.integer': 179}, {24: 496}, ['lval']), ({"'!='": (34, 1, 'expr'), "'>'": (34, 1, 'expr'), 'T.newline': (34, 1, 'expr'), "'='": (34, 1, 'expr'), "'/'": (34, 1, 'expr'), "'-'": (34, 1, 'expr'), "','": (34, 1, 'expr'), "')'": (34, 1, 'expr'), "'<='": (34, 1, 'expr'), "'>='": (34, 1, 'expr'), "'*'": (34, 1, 'expr'), "'}'": (34, 1, 'expr'), "'+'": (34, 1, 'expr'), "'<'": (34, 1, 'expr')}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (42, 1), "'>'": (42, 1), 'T.newline': (42, 1), "'='": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "','": (42, 1), "')'": (42, 1), "'<='": (42, 1), "'>='": (42, 1), "'*'": (42, 1), "'}'": (42, 1), "'+'": (42, 1), "'<'": (42, 1)}, {}, ['expr']), ({"'!='": (53, 1), "'>'": (53, 1), 'T.newline': (53, 1), "'='": (53, 1), "'/'": (53, 1), "'-'": (53, 1), "','": (53, 1), "')'": (53, 1), "'<='": (53, 1), "'>='": (53, 1), "'*'": (53, 1), "'}'": (53, 1), "'+'": (53, 1), "'<'": (53, 1)}, {}, ['expr']), ({"'!='": (54, 1), "'>'": (54, 1), 'T.newline': (54, 1), "'='": (54, 1), "'/'": (54, 1), "'-'": (54, 1), "','": (54, 1), "')'": (54, 1), "'<='": (54, 1), "'>='": (54, 1), "'*'": (54, 1), "'}'": (54, 1), "'+'": (54, 1), "'<'": (54, 1)}, {}, ['expr']), ({"'!='": (55, 1), "'>'": (55, 1), 'T.newline': (55, 1), "'='": (55, 1), "'/'": (55, 1), "'-'": (55, 1), "','": (55, 1), "')'": (55, 1), "'<='": (55, 1), "'>='": (55, 1), "'*'": (55, 1), "'}'": (55, 1), "'+'": (55, 1), "'<'": (55, 1)}, {}, ['expr']), ({"'!='": (56, 1), "'>'": (56, 1), 'T.newline': (56, 1), "'='": (56, 1), "'/'": (56, 1), "'-'": (56, 1), "','": (56, 1), "')'": (56, 1), "'<='": (56, 1), "'>='": (56, 1), "'*'": (56, 1), "'}'": (56, 1), "'+'": (56, 1), "'<'": (56, 1)}, {}, ['expr']), ({"'!='": (57, 1), "'>'": (57, 1), 'T.newline': (57, 1), "'='": (57, 1), "'/'": (57, 1), "'-'": (57, 1), "','": (57, 1), "')'": (57, 1), "'<='": (57, 1), "'>='": (57, 1), "'*'": (57, 1), "'}'": (57, 1), "'+'": (57, 1), "'<'": (57, 1)}, {}, ['expr']), ({"'!='": (58, 1), "'>'": (58, 1), 'T.newline': (58, 1), "'='": (58, 1), "'/'": (58, 1), "'-'": (58, 1), "','": (58, 1), "')'": (58, 1), "'<='": (58, 1), "'>='": (58, 1), "'*'": (58, 1), "'}'": (58, 1), "'+'": (58, 1), "'<'": (58, 1)}, {}, ['expr']), ({"'!='": (59, 1), "'>'": (59, 1), 'T.newline': (59, 1), "'='": (59, 1), "'/'": (59, 1), "'-'": (59, 1), "','": (59, 1), "')'": (59, 1), "'<='": (59, 1), "'>='": (59, 1), "'*'": (59, 1), "'}'": (59, 1), "'+'": (59, 1), "'<'": (59, 1)}, {}, ['expr']), ({'T.integer': 179}, {24: 495}, ['expt']), ({"'min'": 486, 'T.integer': 179, "'max'": 487, "'-'": 485, "'V'": 465, "'('": 178, "'E'": 483, "'+'": 484}, {32: 464, 34: 494, 68: 476, 69: 477, 70: 478, 71: 479, 72: 480, 73: 481, 42: 466, 56: 472, 51: 467, 52: 468, 53: 469, 54: 470, 55: 471, 24: 463, 57: 473, 58: 474, 59: 475, 74: 482}, ['expr']), ({"'min'": 486, 'T.integer': 179, "'max'": 487, "'-'": 485, "'V'": 465, "'('": 178, "'E'": 483, "'+'": 484}, {32: 464, 34: 493, 68: 476, 69: 477, 70: 478, 71: 479, 72: 480, 73: 481, 42: 466, 56: 472, 51: 467, 52: 468, 53: 469, 54: 470, 55: 471, 24: 463, 57: 473, 58: 474, 59: 475, 74: 482}, ['expr']), ({"'('": 492}, {}, ['min']), ({"'('": 491}, {}, ['max']), ({'__end_of_input__': (29, 5, 'then'), "'else'": (29, 5, 'then'), "'}'": (29, 5, 'then'), 'T.newline': (29, 5, 'then')}, {}, ['then']), ({"'{'": 364}, {29: 497}, ['ifs']), ({"'{'": 411}, {29: 498}, ['whiles']), ({"'min'": 523, 'T.integer': 284, "'max'": 524, "'-'": 522, "'V'": 502, "'('": 184, "'E'": 520, "'+'": 521}, {32: 500, 34: 501, 68: 513, 69: 514, 70: 515, 71: 516, 72: 517, 73: 518, 42: 503, 56: 509, 51: 504, 52: 505, 53: 506, 54: 507, 55: 508, 24: 499, 57: 510, 58: 511, 59: 512, 74: 519}, ['max']), ({"'min'": 523, 'T.integer': 284, "'max'": 524, "'-'": 522, "'V'": 502, "'('": 184, "'E'": 520, "'+'": 521}, {32: 500, 34: 525, 68: 513, 69: 514, 70: 515, 71: 516, 72: 517, 73: 518, 42: 503, 56: 509, 51: 504, 52: 505, 53: 506, 54: 507, 55: 508, 24: 499, 57: 510, 58: 511, 59: 512, 74: 519}, ['min']), ({"'!='": (69, 2), "'>'": (69, 2), 'T.newline': (69, 2), "'='": (69, 2), "'/'": (69, 2), "'-'": (69, 2), "','": (69, 2), "')'": (69, 2), "'<='": (69, 2), "'>='": (69, 2), "'*'": (69, 2), "'}'": (69, 2), "'+'": (69, 2), "'<'": (69, 2)}, {}, ['expr']), ({"'!='": (69, 2), "'>'": (69, 2), 'T.newline': (69, 2), "'='": (69, 2), "'/'": (69, 2), "'-'": (69, 2), "','": (69, 2), "')'": (69, 2), "'<='": (69, 2), "'>='": (69, 2), "'*'": (69, 2), "'}'": (69, 2), "'+'": (69, 2), "'<'": (69, 2)}, {}, ['expr']), ({"'!='": (68, 2, 'expt'), "'>'": (68, 2, 'expt'), 'T.newline': (68, 2, 'expt'), "'='": (68, 2, 'expt'), "'/'": (68, 2, 'expt'), "'-'": (68, 2, 'expt'), "','": (68, 2, 'expt'), "')'": (68, 2, 'expt'), "'<='": (68, 2, 'expt'), "'>='": (68, 2, 'expt'), "'*'": (68, 2, 'expt'), "'}'": (68, 2, 'expt'), "'+'": (68, 2, 'expt'), "'<'": (68, 2, 'expt')}, {}, ['expt']), ({"'!='": (32, 2, 'lval'), "'>'": (32, 2, 'lval'), 'T.newline': (32, 2, 'lval'), "'='": (32, 2, 'lval'), "'/'": (32, 2, 'lval'), "'-'": (32, 2, 'lval'), "','": (32, 2, 'lval'), "')'": (32, 2, 'lval'), "'<='": (32, 2, 'lval'), "'>='": (32, 2, 'lval'), "'*'": (32, 2, 'lval'), "'}'": (32, 2, 'lval'), "'+'": (32, 2, 'lval'), "'<'": (32, 2, 'lval')}, {}, ['lval']), ({"'else'": 533, "'}'": (30, 0, 'elses'), 'T.newline': (30, 0, 'elses')}, {30: 532}, ['ifs']), ({"'}'": (20, 5, 'whiles'), 'T.newline': (20, 5, 'whiles')}, {}, ['whiles']), ({"'*'": (51, 1), "'/'": (51, 1), "'-'": (51, 1), "'+'": (51, 1), "','": (51, 1)}, {}, ['expr']), ({"'*'": (52, 1), "'/'": (52, 1), "'-'": (52, 1), "'+'": (52, 1), "','": (52, 1)}, {}, ['expr']), ({"'*'": 316, "'/'": 317, "'-'": 315, "'+'": 314, "','": 534}, {}, ['expr', 'max']), ({'T.integer': 284}, {24: 531}, ['lval']), ({"'*'": (34, 1, 'expr'), "'/'": (34, 1, 'expr'), "'-'": (34, 1, 'expr'), "'+'": (34, 1, 'expr'), "','": (34, 1, 'expr')}, {}, ['expr']), ({"'*'": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "'+'": (42, 1), "','": (42, 1)}, {}, ['expr']), ({"'*'": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "'+'": (42, 1), "','": (42, 1)}, {}, ['expr']), ({"'*'": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "'+'": (42, 1), "','": (42, 1)}, {}, ['expr']), ({"'*'": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "'+'": (42, 1), "','": (42, 1)}, {}, ['expr']), ({"'*'": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "'+'": (42, 1), "','": (42, 1)}, {}, ['expr']), ({"'*'": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "'+'": (42, 1), "','": (42, 1)}, {}, ['expr']), ({"'*'": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "'+'": (42, 1), "','": (42, 1)}, {}, ['expr']), ({"'*'": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "'+'": (42, 1), "','": (42, 1)}, {}, ['expr']), ({"'*'": (42, 1), "'/'": (42, 1), "'-'": (42, 1), "'+'": (42, 1), "','": (42, 1)}, {}, ['expr']), ({"'*'": (53, 1), "'/'": (53, 1), "'-'": (53, 1), "'+'": (53, 1), "','": (53, 1)}, {}, ['expr']), ({"'*'": (54, 1), "'/'": (54, 1), "'-'": (54, 1), "'+'": (54, 1), "','": (54, 1)}, {}, ['expr']), ({"'*'": (55, 1), "'/'": (55, 1), "'-'": (55, 1), "'+'": (55, 1), "','": (55, 1)}, {}, ['expr']), ({"'*'": (56, 1), "'/'": (56, 1), "'-'": (56, 1), "'+'": (56, 1), "','": (56, 1)}, {}, ['expr']), ({"'*'": (57, 1), "'/'": (57, 1), "'-'": (57, 1), "'+'": (57, 1), "','": (57, 1)}, {}, ['expr']), ({"'*'": (58, 1), "'/'": (58, 1), "'-'": (58, 1), "'+'": (58, 1), "','": (58, 1)}, {}, ['expr']), ({"'*'": (59, 1), "'/'": (59, 1), "'-'": (59, 1), "'+'": (59, 1), "','": (59, 1)}, {}, ['expr']), ({'T.integer': 284}, {24: 530}, ['expt']), ({"'min'": 208, 'T.integer': 284, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 115, "'E'": 205, "'+'": 206}, {32: 185, 34: 529, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['expr']), ({"'min'": 208, 'T.integer': 284, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 115, "'E'": 205, "'+'": 206}, {32: 185, 34: 528, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['expr']), ({"'('": 527}, {}, ['min']), ({"'('": 526}, {}, ['max']), ({"'*'": 316, "'/'": 317, "'-'": 315, "'+'": 314, "','": 535}, {}, ['expr', 'min']), ({"'min'": 208, 'T.integer': 284, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 115, "'E'": 205, "'+'": 206}, {32: 185, 34: 116, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['max']), ({"'min'": 208, 'T.integer': 284, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 115, "'E'": 205, "'+'": 206}, {32: 185, 34: 286, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['min']), ({"'*'": (69, 2), "'/'": (69, 2), "'-'": (69, 2), "'+'": (69, 2), "','": (69, 2)}, {}, ['expr']), ({"'*'": (69, 2), "'/'": (69, 2), "'-'": (69, 2), "'+'": (69, 2), "','": (69, 2)}, {}, ['expr']), ({"'*'": (68, 2, 'expt'), "'/'": (68, 2, 'expt'), "'-'": (68, 2, 'expt'), "'+'": (68, 2, 'expt'), "','": (68, 2, 'expt')}, {}, ['expt']), ({"'*'": (32, 2, 'lval'), "'/'": (32, 2, 'lval'), "'-'": (32, 2, 'lval'), "'+'": (32, 2, 'lval'), "','": (32, 2, 'lval')}, {}, ['lval']), ({"'}'": (19, 6, 'ifs'), 'T.newline': (19, 6, 'ifs')}, {}, ['ifs']), ({"'{'": 364}, {29: 537}, ['elses']), ({"'min'": 208, 'T.integer': 117, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 184, "'E'": 205, "'+'": 206}, {32: 185, 34: 538, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['max']), ({"'min'": 208, 'T.integer': 117, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 184, "'E'": 205, "'+'": 206}, {32: 185, 34: 539, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['min']), ({"'min'": 208, 'T.integer': 117, "'max'": 209, "'-'": 207, "'V'": 187, "'('": 184, "'E'": 205, "'+'": 206}, {32: 185, 34: 540, 68: 198, 69: 199, 70: 200, 71: 201, 72: 202, 73: 203, 42: 188, 56: 194, 51: 189, 52: 190, 53: 191, 54: 192, 55: 193, 24: 183, 57: 195, 58: 196, 59: 197, 74: 204}, ['max']), ({"'}'": (30, 2, 'elses'), 'T.newline': (30, 2, 'elses')}, {}, ['elses']), ({"')'": 541, "'*'": 316, "'/'": 317, "'-'": 315, "'+'": 314}, {}, ['expr', 'max']), ({"')'": 542, "'*'": 316, "'/'": 317, "'-'": 315, "'+'": 314}, {}, ['expr', 'min']), ({"')'": 543, "'*'": 316, "'/'": 317, "'-'": 315, "'+'": 314}, {}, ['expr', 'max']), ({"'!='": (74, 6, 'max'), "'>'": (74, 6, 'max'), 'T.newline': (74, 6, 'max'), "'='": (74, 6, 'max'), "'/'": (74, 6, 'max'), "'-'": (74, 6, 'max'), "','": (74, 6, 'max'), "')'": (74, 6, 'max'), "'<='": (74, 6, 'max'), "'>='": (74, 6, 'max'), "'*'": (74, 6, 'max'), "'}'": (74, 6, 'max'), "'+'": (74, 6, 'max'), "'<'": (74, 6, 'max')}, {}, ['max']), ({"'!='": (73, 6, 'min'), "'>'": (73, 6, 'min'), 'T.newline': (73, 6, 'min'), "'='": (73, 6, 'min'), "'/'": (73, 6, 'min'), "'-'": (73, 6, 'min'), "','": (73, 6, 'min'), "')'": (73, 6, 'min'), "'<='": (73, 6, 'min'), "'>='": (73, 6, 'min'), "'*'": (73, 6, 'min'), "'}'": (73, 6, 'min'), "'+'": (73, 6, 'min'), "'<'": (73, 6, 'min')}, {}, ['min']), ({"'!='": (74, 6, 'max'), "'>'": (74, 6, 'max'), 'T.newline': (74, 6, 'max'), '__end_of_input__': (74, 6, 'max'), "'/'": (74, 6, 'max'), "'-'": (74, 6, 'max'), "','": (74, 6, 'max'), "')'": (74, 6, 'max'), "'<='": (74, 6, 'max'), "'>='": (74, 6, 'max'), "'*'": (74, 6, 'max'), "'='": (74, 6, 'max'), "'+'": (74, 6, 'max'), "'<'": (74, 6, 'max')}, {}, ['max']))
